'''
Functions used to compute earthquake focal mechanisms.

Significant portions of the functions in this file are based on the Fortran HASH
code originally written by Jeanne L. Hardebeck & Peter M. Shearer, and all of it
is inspired by their work. Please cite the appropriate references if you use
this code.
'''

# Standard libraries
import os

# External libraries
import numpy as np
import pandas as pd
rng=np.random.default_rng(123) # Used to produce reproducable bootstrapped results.


def lookup_takeoff(table,perturbed_origin_depth_km,sr_dist_km,look_dep,look_del,deptab,delttab,num_velocity_models=1):
	'''
	Given a hypocentral depths and source-receiver distances, queries the takeoff angles
	using the lookup table
	Input:
		table: Lookup table produced by create_takeoff_table(), 3d array.
		perturbed_origin_depth_km: perturbed origin depths in km produced by perturb_eq_locations, 2d array.
		sr_dist_km: source-receiver distances in km produced by perturb_eq_locations, 2d array.
		look_dep: minimum source depth, maximum, and interval for the lookup table, list
		look_del: minimum source-station distance, maximum, and interval for the lookup table, list
		deptab: array of depths (km) produced by create_lookup_table()
		delttab: array of distances (km) produced by create_lookup_table()
		num_velocity_models: the number of velocity models used.
	Output:
		takeoff: the corresponding takeoff angles
	'''
	# Determines the lookup table cell for permiated earthquake depths
	id1=((perturbed_origin_depth_km-look_dep[0])/look_dep[2]).astype(int)
	id2=id1+1

	# If perturbed locations are outside of the modeled range, sets those perturbations to the model edge
	dist_flag=sr_dist_km<delttab[0]
	if np.any(dist_flag):
		print('*WARNING: Perturbed epicenters(s) are closer than lookup table depth range. Setting these perturbed locations to {} km *'.format(delttab[0]))
		sr_dist_km[dist_flag]=delttab[0]
	dist_flag=sr_dist_km>delttab[-1]
	if np.any(dist_flag):
		print('*WARNING: Perturbed epicenters(s) are further than lookup table depth range. Setting these perturbed locations to {} km *'.format(delttab[-1]))
		sr_dist_km[dist_flag]=delttab[-1]

	# Determines the lookup table cell for permiated earthquake epicenters
	ix1=((sr_dist_km-look_del[0])/look_del[2]).astype(int)
	ix2=ix1+1

	# Determines which velocity model to use for each trial
	if num_velocity_models>1:
		randomized_vm_ind=rng.integers(low=0, high=num_velocity_models,size=perturbed_origin_depth_km.shape[-1])
		randomized_vm_ind[0]=0
	else:
		randomized_vm_ind=np.zeros(perturbed_origin_depth_km.shape[-1],dtype=int)

	# Uses the lookup table to determine the travel times / takeoff angles
	xfrac=(sr_dist_km-delttab[ix1])/(delttab[ix2]-delttab[ix1])

	if id1.ndim==1:
		id1=np.repeat(id1[np.newaxis,:],len(sr_dist_km),axis=0)
		id2=np.repeat(id2[np.newaxis,:],len(sr_dist_km),axis=0)

	t1=table[ix1,id1,np.repeat([randomized_vm_ind],len(sr_dist_km),axis=0)]+\
		xfrac*(table[ix2,id1,np.repeat([randomized_vm_ind],len(sr_dist_km),axis=0)]-
		table[ix1,id1,np.repeat([randomized_vm_ind],len(sr_dist_km),axis=0)])

	t2=table[ix1,id2,np.repeat([randomized_vm_ind],len(sr_dist_km),axis=0)]+\
		xfrac*(table[ix2,id2,np.repeat([randomized_vm_ind],len(sr_dist_km),axis=0)]-
		table[ix1,id2,np.repeat([randomized_vm_ind],len(sr_dist_km),axis=0)])

	dfrac=(perturbed_origin_depth_km-deptab[id1])/(deptab[id2]-deptab[id1])
	takeoff=t1+dfrac*(t2-t1)

	return takeoff


def perturb_eq_locations(event_pol_df,look_dep,perturb_epicentral_location,nmc=1):
	'''
	Randomly perturbs the hypocentral locations.
	Input:
		event_pol_df: polarity dataframe
		look_dep: minimum source depth, maximum, and interval for the lookup table, list
		perturb_epicentral_location: flag to determine if epicentral locations should be perturbed, boolean
		nmc: number of trials, integer
	Output:
		perturbed_origin_depth_km: Perturbed hypocentral depths
		sr_dist_km: Perturbed source-receiver distances
		sr_azimuth: Perturbed source-receiver azimuths
	'''
	if 'event_id2' in event_pol_df:
		unique_event_id2,unique_ind,unique_inv=np.unique(event_pol_df.event_id2.values,return_index=True,return_inverse=True)
		num_unique_events=len(unique_event_id2)
	else:
		num_unique_events=1
		unique_ind=[0]
		unique_inv=np.zeros(len(event_pol_df),dtype=int)

	rng_z=rng.normal(size=(nmc,num_unique_events))

	rng_z[0]=0
	perturbed_origin_depth_km=event_pol_df.iloc[unique_ind]['origin_depth_km'].values+rng_z*event_pol_df.iloc[unique_ind]['vert_uncert_km'].values
	perturbed_origin_depth_km=perturbed_origin_depth_km[:,unique_inv]

	if perturb_epicentral_location: # Perturbs horizontal earthquake locations
		perturb_rand_azimuth=rng.integers(low=0, high=360, size=(nmc,num_unique_events))
		rng_h=rng.normal(size=(nmc,num_unique_events))
		rng_h[0]=0

		# Earthquake lon/lat in deg convertered to radians
		eq_lon_r=np.deg2rad(event_pol_df.iloc[unique_ind]['origin_lon'].values)
		eq_lat_r=np.deg2rad(event_pol_df.iloc[unique_ind]['origin_lat'].values)
		perturb_rand_horz_km=event_pol_df.iloc[unique_ind]['horz_uncert_km'].values*rng_h

		earth_radius_km=6371 # approx radius of earth in km
		eq_pert_lat_r=np.arcsin(np.sin(eq_lat_r)*np.cos(perturb_rand_horz_km/earth_radius_km)+
									np.cos(eq_lat_r)*np.sin(perturb_rand_horz_km/earth_radius_km)*np.cos(perturb_rand_azimuth))
		eq_pert_lon_r=eq_lon_r+np.arctan2(np.sin(perturb_rand_azimuth)*np.sin(perturb_rand_horz_km/earth_radius_km)*np.cos(eq_lat_r),
					np.cos(perturb_rand_horz_km/earth_radius_km)-np.sin(eq_lat_r)*np.sin(eq_pert_lat_r))

		aspect=np.cos(np.median(eq_pert_lat_r[:,0]))
		qlon=np.rad2deg(eq_pert_lon_r)
		qlat=np.rad2deg(eq_pert_lat_r)

	else:
		qlon=np.atleast_2d(event_pol_df.iloc[unique_ind]['origin_lon'].values)
		qlat=np.atleast_2d(event_pol_df.iloc[unique_ind]['origin_lat'].values)
		aspect=np.cos(np.deg2rad(np.median(qlat)))

	qlon=qlon[:,unique_inv]
	qlat=qlat[:,unique_inv]

	# Ensures all perturbed depths are within the modeled range
	qdep_v_flag=perturbed_origin_depth_km<look_dep[0]
	if np.any(qdep_v_flag):
		print('*WARNING: Perturbed locations for event_id \'{}\' are shallower than lookup table depth range ({}). Setting these perturbed depths to {} km.'.format(event_pol_df.iloc[unique_ind[0]]['event_id'],look_dep[0],look_dep[0]))
		perturbed_origin_depth_km[qdep_v_flag]=look_dep[0]

	qdep_v_flag=perturbed_origin_depth_km>look_dep[1]
	if np.any(qdep_v_flag):
		print('*WARNING: Perturbed locations for event_id \'{}\' are deeper than lookup table depth range ({}). Setting these perturbed depths to {} km.'.format(event_pol_df.iloc[unique_ind[0]]['event_id'],look_dep[1],look_dep[1]))
		perturbed_origin_depth_km[qdep_v_flag]=look_dep[1]-.01

	# Approximates the source-receiver distances and azimuths
	flon=event_pol_df['station_lon'].values
	flat=event_pol_df['station_lat'].values
	dx=(flon-qlon)*111.2*aspect
	dy=(flat-qlat)*111.2
	sr_dist_km=np.sqrt(dx**2+dy**2)
	sr_azimuth=90.-np.rad2deg(np.arctan2(dy,dx))
	sr_azimuth[sr_azimuth<0]+=360

	return perturbed_origin_depth_km.T,sr_dist_km.T,sr_azimuth.T


def perturb_azimuth_takeoff(event_pol_df,nmc):
	'''
	Perturb predetermined azimuth and takeoff angles using the given uncertainties.
	Input:
		event_pol_df: polarity dataframe
		nmc: number of trials, integer
	Output:
		sr_azimuth: perturbed source-receiver azimuths
		takeoff: perturbed takeoff angles
	'''
	rng_az=rng.normal(size=(len(event_pol_df),nmc))
	rng_az[:,0]=0

	rng_takeoff=rng.normal(size=(len(event_pol_df),nmc))
	rng_takeoff[:,0]=0

	sr_azimuth=event_pol_df['azimuth'].values[:,np.newaxis]+event_pol_df['azimuth_uncertainty'].values[:,np.newaxis]*rng_az
	sr_azimuth[sr_azimuth<0]+=360
	sr_azimuth[sr_azimuth>360]-=360

	takeoff=event_pol_df['takeoff'].values[:,np.newaxis]+event_pol_df['takeoff_uncertainty'].values[:,np.newaxis]*rng_takeoff
	takeoff[takeoff<=0]=1e-6
	takeoff[takeoff>180]=180

	return sr_azimuth,takeoff


def cartesian_transform(the,phi,r):
	'''
	Transforms spherical coordinates to cartesian (theta in degrees)
	'''
	x=r*np.sin(np.deg2rad(the))*np.cos(np.deg2rad(phi))
	y=r*np.sin(np.deg2rad(the))*np.sin(np.deg2rad(phi))
	z=-r*np.cos(np.deg2rad(the))
	return x,y,z


def vector_from_sdr(strike_r,dip_r,rake_r):
	'''
	Gets fault normal vector (fnorm) and slip vector (slip) from [strike,dip,rake] (in radians).
	Uses (x,y,z) coordinate system with x=north, y=east, z=down
		Reference:  Aki and Richards, p. 115
	Based on code from HASH (Hardebeck & Shearer, 2002).
	'''
	fnorm=np.zeros(3)
	slip=np.zeros(3)

	fnorm[0]=-np.sin(dip_r)*np.sin(strike_r)
	fnorm[1]=np.sin(dip_r)*np.cos(strike_r)
	fnorm[2]=-np.cos(dip_r)
	slip[0]=np.cos(rake_r)*np.cos(strike_r)+np.cos(dip_r)*np.sin(rake_r)*np.sin(strike_r)
	slip[1]=np.cos(rake_r)*np.sin(strike_r)-np.cos(dip_r)*np.sin(rake_r)*np.cos(strike_r)
	slip[2]=-np.sin(rake_r)*np.sin(dip_r)

	return fnorm,slip


def sdr_from_vector(faultnorms,slips):
	'''
	Gets [strike,dip,rake] from fault normal vectors (faultnorms) and slip vector (slips).
	faultnorms and slips should each be a 3-by-n array.

	Uses (x,y,z) coordinate system with x=north, y=east, z=down
		Reference:  Aki and Richards, p. 115
	Based on code from HASH (Hardebeck & Shearer, 2002).
	'''
	num_vect=faultnorms.shape[1]
	if faultnorms.shape != slips.shape:
		raise ValueError('***Error: shape of faultnorms and slips must be the same')
	if faultnorms.shape[0]!=3:
		raise ValueError('***Error: faultnorms and slips must be an array of shape 3-by-n')

	phi=np.zeros(num_vect)
	delt=np.zeros(num_vect)
	lam=np.zeros(num_vect)

	undef_flag=(1-np.abs(faultnorms[2,:]))<=(1e-7)
	if np.any(undef_flag):
		undef_ind=np.where(undef_flag)[0]
		print('*sdr_from_vector warning, horz fault, strike undefined')
		phi[undef_ind]=np.arctan2(-slips[0,undef_ind],slips[1,undef_ind])
		clam=np.cos(phi[undef_ind])*slips[0,undef_ind]+np.sin(phi[undef_ind])*slips[1,undef_ind]
		slam=np.sin(phi[undef_ind])*slips[0,undef_ind]-np.cos(phi[undef_ind])*slips[1,undef_ind]
		lam[undef_ind]=np.arctan2(slam,clam)
	if np.any(~undef_flag):
		def_ind=np.where(~undef_flag)[0]

		phi[def_ind]=np.arctan2(-faultnorms[0,def_ind],faultnorms[1,def_ind])
		a=np.sqrt(faultnorms[0,def_ind]*faultnorms[0,def_ind]+faultnorms[1,def_ind]*faultnorms[1,def_ind])
		delt[def_ind]=np.arctan2(a,-faultnorms[2,def_ind])
		clam=np.cos(phi[def_ind])*slips[0,def_ind]+np.sin(phi[def_ind])*slips[1,def_ind]
		slam=-slips[2,def_ind]/np.sin(delt[def_ind])
		lam[def_ind]=np.arctan2(slam,clam)

		tmp_ind=def_ind[np.where(delt[def_ind]>(0.5*np.pi))[0]]
		if len(tmp_ind)>0:
			delt[tmp_ind]=np.pi-delt[tmp_ind]
			phi[tmp_ind]=phi[tmp_ind]+np.pi
			lam[tmp_ind]=-lam[tmp_ind]

	strike=np.rad2deg(phi)
	dip=np.rad2deg(delt)
	rake=np.rad2deg(lam)

	strike[strike<0]+=360
	rake[rake<-180]+=360
	rake[rake>180]-=360

	return strike,dip,rake


def create_takeoff_table(vmodel_depthvp,deptab,delttab,nump,nx0,nd0,takeoff_az_precision):
	'''
	Creates tables of takeoff angles given a 1D velocity model.
	Based on code from HASH (Hardebeck & Shearer, 2002).
	Input:
		vmodel_depthvp: depth (km) and Vp (km/s), 2d array
		deptab: deptab: array of depths (km)
		delttab: array of distances (km)
		nump: number of rays traced
		nx0: maximum source-station distance bins for look-up tables
		nd0: maximum source depth bins for look-up tables
		takeoff_az_precision: number of decimal places to round resulting table
	Output:
		table: produced lookup table of takeoff angles
	'''
	if vmodel_depthvp[:,0][-1]<deptab[-1]:
		vmodel_depthvp=np.vstack((vmodel_depthvp,vmodel_depthvp[-1,:]))
		vmodel_depthvp[-1,0]=deptab[-1]+1
		vmodel_depthvp[-1,1]=vmodel_depthvp[-1,1]+0.001

	table=np.zeros((nx0,len(deptab)))-999

	ndel=len(delttab)
	ndep=len(deptab)
	pmin=0

	z=vmodel_depthvp[:,0]
	alpha=vmodel_depthvp[:,1]

	npts=len(z)
	z=np.hstack((z,z[npts-1]))
	alpha=np.hstack((alpha,alpha[npts-1]))

	for i in range(npts-1,0,-1):
		for idep in range(ndep-1,-1,-1):
			if (z[i-1]<=(deptab[idep]-0.00001)) & (z[i]>=(deptab[idep]+.00001)):
				z=np.insert(z,i,z[i-1])
				alpha=np.insert(alpha,i,alpha[i-1])
				z[i]=deptab[idep]
				frac=(z[i]-z[i-1])/(z[i+1]-z[i-1])
				alpha[i]=alpha[i-1]+frac*(alpha[i+1]-alpha[i-1])

	slow=1/alpha
	pmax=slow[0]
	pstep=(pmax-pmin)/nump

	# do P-wave ray tracing
	npmax=int((pmax+pstep/2-pmin)/pstep)+1

	depxcor=np.zeros((npmax,nd0))
	depucor=np.zeros((npmax,nd0))
	deptcor=np.zeros((npmax,nd0))

	tmp_ind=np.where(deptab==0)[0]

	tmp_ind=np.where(deptab!=0)[0]
	depxcor[:,tmp_ind]=-999
	deptcor[:,tmp_ind]=-999

	ptab=np.linspace(pmin,pmin+pstep*(npmax-1),num=npmax)

	h_array=z[1:]-z[:-1]
	utop=slow[:-1]
	ubot=slow[1:]

	'''LAYERTRACE equivalent'''
	dx=np.zeros((npmax,len(utop)))
	dt=np.zeros((npmax,len(utop)))
	irtr=np.zeros((npmax,len(utop)),dtype=int)

	qs=np.zeros((npmax,len(utop)));qs[:]=np.nan
	qr=np.zeros((npmax,len(utop)));qr[:]=np.nan
	ytop=utop-ptab[:, np.newaxis]
	ytop_pos_flag=ytop>0
	qs[ytop_pos_flag]=ytop[ytop_pos_flag]*(utop+ptab[:, np.newaxis])[ytop_pos_flag]
	qs[ytop_pos_flag]=np.sqrt(qs[ytop_pos_flag])

	qr=np.arctan2(qs,ptab[:, np.newaxis])

	b=np.ma.divide(-np.log(ubot/utop),h_array)
	b=b.filled(np.nan)

	# integral at upper limit, 1/b factor omitted until end
	etau=qs-qr*ptab[:, np.newaxis]
	ex=qr

	# check lower limit to see if we have turning point
	ybot=ubot-ptab[:, np.newaxis]
	# if turning point, then no contribution from bottom point
	y_subzero_flag=(ybot<=0)
	y_greaterzero_flag=(ybot>0)
	irtr[y_subzero_flag ]=2
	irtr[y_greaterzero_flag]=1

	irtr[~ytop_pos_flag]=0

	dx[y_subzero_flag]=ex[y_subzero_flag]
	dx=dx/b
	dtau=etau/b
	dt[y_subzero_flag]=dtau[y_subzero_flag]+(dx*ptab[:, np.newaxis])[y_subzero_flag] # converts tau to t

	q=np.zeros(ybot.shape);q[:]=np.nan
	q[y_greaterzero_flag]=ybot[y_greaterzero_flag]*((ubot+ptab[:, np.newaxis])[y_greaterzero_flag])
	qs=np.sqrt(q)

	qr=np.arctan2(qs,ptab[:,np.newaxis])
	etau=etau-qs+ptab[:, np.newaxis]*qr
	ex=ex-qr

	exb=ex/b
	dtau=etau/b
	dx[y_greaterzero_flag]=exb[y_greaterzero_flag]
	dt[y_greaterzero_flag]=dtau[y_greaterzero_flag]+(exb*ptab[:, np.newaxis])[y_greaterzero_flag]

	# Ensures values after ray has turned are nan
	x=((irtr==0) | (irtr==2))
	idx=np.arange(npmax), x.argmax(axis=1)
	tmp=(x[idx]==True)

	idx_1=idx[0][tmp],idx[1][tmp]+1
	tmp_1=idx_1[1]<(len(utop)-1)
	idx_1=idx_1[0][tmp_1],idx_1[1][tmp_1]
	for xx in range(len(idx_1[0])):
		row=idx_1[0][xx]
		col=idx_1[1][xx]
		dx[row,col:]=np.nan
		dt[row,col:]=np.nan
	deltab=np.nansum(dx,axis=1)*2
	tttab=np.nansum(dt,axis=1)*2

	idx_2=idx[0][tmp],idx[1][tmp]
	tmp_2=idx_2[1]<(len(utop)-1)
	idx_2=idx_2[0][tmp_2],idx_2[1][tmp_2]
	for xx in range(len(idx_2[0])):
		row=idx_2[0][xx]
		col=idx_2[1][xx]
		dx[row,col:]=np.nan
		dt[row,col:]=np.nan

	depxcor=np.cumsum(dx,axis=1)
	deptcor=np.cumsum(dt,axis=1)
	output_col_ind=np.where(np.isin(z,deptab))[0]-1
	depxcor=depxcor[:,output_col_ind]
	deptcor=deptcor[:,output_col_ind]
	depxcor[:,0]=0
	deptcor[:,0]=0
	depucor[:]=slow[output_col_ind+1]
	depucor[np.isnan(depxcor)]=-999
	depxcor[np.isnan(depxcor)]=-999
	deptcor[np.isnan(deptcor)]=-999

	x=np.diff(depxcor,axis=0)<=0
	idx=x.argmax(axis=0)+1
	tmp=(x[idx,np.arange(nd0)]==False)
	idx[tmp]=npmax-1

	for idep in range(ndep):
		# upgoing rays from source
		xsave_up=depxcor[:(idx[idep]),idep]
		tsave_up=deptcor[:(idx[idep]),idep]
		usave_up=depucor[:(idx[idep]),idep]
		psave_up=-1*ptab[:(idx[idep])]

		# downgoing rays from source
		down_idx=np.where((depxcor[:,idep]!=-999) & (deltab!=-999))[0][::-1]
		xsave_down=(deltab[down_idx]-depxcor[down_idx,idep])
		tsave_down=(tttab[down_idx]-deptcor[down_idx,idep])
		usave_down=depucor[down_idx,idep]
		psave_down=ptab[down_idx]

		# Merges upgoing and downgoing ray arrays
		xsave=np.hstack([xsave_up,xsave_down])
		tsave=np.hstack([tsave_up,tsave_down])
		usave=np.hstack([usave_up,usave_down])
		psave=np.hstack([psave_up,psave_down])

		scr1=np.zeros(ndel)
		for idel in range(1,ndel):
			del_x=delttab[idel]
			ind=np.where( (xsave[:-1]<=del_x) & (xsave[1:]>=del_x) )[0]+1

			frac=(del_x-xsave[ind-1])/(xsave[ind]-xsave[ind-1])
			t1=tsave[ind-1]+frac*(tsave[ind]-tsave[ind-1])

			min_ind=ind[np.argmin(t1)]

			scr1[idel]=psave[min_ind]/usave[min_ind]
		angle=np.rad2deg(np.arcsin(scr1))

		angle_flag=angle>=0
		angle*=-1
		angle[angle_flag]+=180

		table[:,idep]=angle

	if delttab[0]==0:
		table[0,:]=0. # straight up at zero range
	table=np.round(table,takeoff_az_precision)
	return table


def dir_cos_setup(p_dict):
	'''
	Sets up array with direction cosines for all coordinate transformations
	Input:
		p_dict: Parameter values created in SKHASH.py, dictionary
	Output:
		dir_cos_dict: Coordinate transformation variables, dictionary
	'''
	ntab=180
	astep=1/ntab

	dir_cos_dict={
		'ncoor':0,
		'thetable':np.empty(0),
		'phitable':np.empty(0),
		'b1':np.empty(0),
		'b2':np.empty(0),
		'b3':np.empty(0),
		'amptable':np.empty(0)
	}

	num_izeta=int(np.floor(179.9/p_dict['dang']))
	the=np.arange(0,90.001,p_dict['dang']) # angles for grid search

	rthe=np.deg2rad(the)
	costhe=np.cos(rthe)
	sinthe=np.sin(rthe)
	fnumang=360./p_dict['dang']
	dphi=np.round(fnumang*np.sin(rthe))
	dphi[dphi!=0]=360./dphi[dphi!=0]
	dphi[dphi==0]=10000.
	num_iphi=np.floor(359.9/dphi).astype(int)

	dir_cos_dict['ncoor']=np.sum(num_iphi+1)*(num_izeta+1)

	dir_cos_dict['b1']=np.zeros((3,dir_cos_dict['ncoor']))
	dir_cos_dict['b2']=np.zeros((3,dir_cos_dict['ncoor']))
	dir_cos_dict['b3']=np.zeros((3,dir_cos_dict['ncoor']))
	irot=0

	for ithe in range(len(the)):
		bb1=np.zeros(3)
		bb3=np.zeros(3)
		for iphi in range(0,num_iphi[ithe]+1):
			phi=iphi*dphi[ithe]
			rphi=np.deg2rad(phi)
			cosphi=np.cos(rphi)
			sinphi=np.sin(rphi)
			bb3[2]=costhe[ithe]
			bb3[0]=sinthe[ithe]*cosphi
			bb3[1]=sinthe[ithe]*sinphi

			bb1[2]=-sinthe[ithe]
			bb1[0]=costhe[ithe]*cosphi
			bb1[1]=costhe[ithe]*sinphi

			bb2=np.cross(bb1,bb3)*-1

			for izeta in range(0,num_izeta+1):
				if (irot>dir_cos_dict['ncoor']):
					raise ValueError('***FOCAL error: # of rotations too big')
				zeta=izeta*p_dict['dang']
				rzeta=np.deg2rad(zeta)
				coszeta=np.cos(rzeta)
				sinzeta=np.sin(rzeta)

				dir_cos_dict['b3'][2,irot]=bb3[2]
				dir_cos_dict['b3'][0,irot]=bb3[0]
				dir_cos_dict['b3'][1,irot]=bb3[1]
				dir_cos_dict['b1'][0,irot]=bb1[0]*coszeta+bb2[0]*sinzeta
				dir_cos_dict['b1'][1,irot]=bb1[1]*coszeta+bb2[1]*sinzeta
				dir_cos_dict['b1'][2,irot]=bb1[2]*coszeta+bb2[2]*sinzeta
				dir_cos_dict['b2'][0,irot]=bb2[0]*coszeta-bb1[0]*sinzeta
				dir_cos_dict['b2'][1,irot]=bb2[1]*coszeta-bb1[1]*sinzeta
				dir_cos_dict['b2'][2,irot]=bb2[2]*coszeta-bb1[2]*sinzeta
				irot=irot+1

	bbb=-1.+np.arange(0,2*ntab+1)*astep
	_x,_y=np.meshgrid(bbb,bbb)
	dir_cos_dict['thetable']=np.arccos(bbb)
	dir_cos_dict['phitable']=np.arctan2(_x,_y).T
	dir_cos_dict['phitable'][dir_cos_dict['phitable']<0]+=2*np.pi

	# Creates amptable
	if p_dict['ampfile']:
		dir_cos_dict['amptable']=np.zeros((2,ntab,2*ntab))
		phi=np.arange(0,2*ntab)*np.pi*astep
		theta=np.arange(0,ntab)*np.pi*astep

		_x,_y=np.meshgrid(theta,phi)
		dir_cos_dict['amptable'][0,:,:]=np.abs(np.sin(2*_x)*np.cos(_y)).T

		s1=np.cos(2*_x)*np.cos(_y)
		s2=-np.cos(_x)*np.sin(_y)
		dir_cos_dict['amptable'][1,:,:]=np.sqrt(s1*s1+s2*s2).T

		if p_dict['min_amp']>0:
			dir_cos_dict['amptable'][dir_cos_dict['amptable']<p_dict['min_amp']]=0.0

	return dir_cos_dict


def average_mech(norm1in,norm2in):
	'''
	Computes the average mech of the solutions.

	Inputs:
		norm1in: normal to fault plane, array(3,nf)
		norm2in: slip vector, array(3,nf)
	Output:
		norm1_avg: normal to average of plane 1
		norm2_avg: normal to average of plane 2
	'''
	if norm1in.shape != norm2in.shape:
		raise ValueError('***Error in average_mech: shape of norm1in and norm2in must be the same')
	if len(norm1in.shape)!=2:
		raise ValueError('***Error in average_mech: norm1in and norm2in must each be an array of shape n-by-3')
	if norm1in.shape[0]!=3:
		raise ValueError('***Error in average_mech: norm1in and norm2in must each be an array of shape n-by-3')

	norm1=norm1in.copy()
	norm2=norm2in.copy()

	# If there is only one mechanism, return that mechanism
	if norm1.shape[1]==1:
		return norm1[:,0],norm2[:,0]

	norm1_ref=norm1in[:,0].copy()
	norm2_ref=norm2in[:,0].copy()

	rota,temp1,temp2=mech_rotation(norm1_ref,norm1[:,1:],norm2_ref,norm2[:,1:])

	norm1_avg=np.sum(np.hstack((norm1[:,[0]],temp1)),axis=1)
	norm2_avg=np.sum(np.hstack((norm2[:,[0]],temp2)),axis=1)
	ln_norm1=np.sqrt(np.sum(norm1_avg**2))
	ln_norm2=np.sqrt(np.sum(norm2_avg**2))
	norm1_avg=norm1_avg/ln_norm1
	norm2_avg=norm2_avg/ln_norm2

	# Determine the RMS observed angular difference between the average
	# Normal vectors and the normal vectors of each mechanism
	rota,temp1,temp2=mech_rotation(norm1_avg,norm1,norm2_avg,norm2)
	d11=temp1[0,:]*norm1_avg[0]+temp1[1,:]*norm1_avg[1]+temp1[2,:]*norm1_avg[2]
	d22=temp2[0,:]*norm2_avg[0]+temp2[1,:]*norm2_avg[1]+temp2[2,:]*norm2_avg[2]

	d11[d11>1]=1
	d11[d11<-1]=-1
	d22[d22>1]=1
	d22[d22<-1]=-1

	a11=np.arccos(d11)
	a22=np.arccos(d22)

	avang1=np.sqrt(np.sum(a11**2)/len(a11))
	avang2=np.sqrt(np.sum(a22**2)/len(a22))

	# the average normal vectors may not be exactly orthogonal (although
	# usually they are very close) - find the misfit from orthogonal and
	# adjust the vectors to make them orthogonal - adjust the more poorly
	# constrained plane more
	if (avang1+avang2)>=0.0001:
		maxmisf=0.01
		fract1=avang1/(avang1+avang2)
		for icount in range(100):
			dot1=norm1_avg[0]*norm2_avg[0]+norm1_avg[1]*norm2_avg[1]+norm1_avg[2]*norm2_avg[2]
			misf=90-np.rad2deg(np.arccos(dot1))
			if abs(misf)<=maxmisf:
				break
			else:
				theta1=np.deg2rad(misf*fract1)
				theta2=np.deg2rad(misf*(1-fract1))
				temp=norm1_avg
				norm1_avg=norm1_avg-norm2_avg*np.sin(theta1)
				norm2_avg=norm2_avg-temp*np.sin(theta2)
				ln_norm1=np.sqrt(np.sum(norm1_avg*norm1_avg))
				ln_norm2=np.sqrt(np.sum(norm2_avg*norm2_avg))
				norm1_avg=norm1_avg/ln_norm1
				norm2_avg=norm2_avg/ln_norm2
	return norm1_avg,norm2_avg


def mech_rotation(norm1_in,norm2_in,slip1_in,slip2_in):
	'''
	Finds the minimum rotation angle between two mechanisms.
	Does not assume that the normal and slip vectors are matched.
	Input:
		norm1_in: normal to fault plane 1
		norm2_in: normal to fault plane 2
		slip1_in: slip vector 1
		slip2_in: slip vector 2
	Output:
		rota: rotation angle
		norm2: normal to fault plane, best combination
		slip2: slip vector, best combination
	'''
	if norm1_in.shape != slip1_in.shape:
		raise ValueError('***Error in mech_rotation: shape of norm1_in and slip1_in must be the same')
	if norm2_in.shape != slip2_in.shape:
		raise ValueError('***Error in mech_rotation: shape of norm2_in and slip2_in must be the same')
	if len(norm1_in.shape)!=1:
		raise ValueError('***Error in mech_rotation: norm1_in and slip1_in must each be an array of length 3')
	if norm1_in.shape[0]!=3:
		raise ValueError('***Error in mech_rotation: norm1_in and slip1_in must each be an array of length 3')
	if len(norm2_in.shape)!=2:
		raise ValueError('***Error in mech_rotation: norm2_in and slip2_in must each be an array of shape 3-by-n')
	if norm2_in.shape[0]!=3:
		raise ValueError('***Error in mech_rotation: norm2_in and slip2_in must each be an array of shape 3-by-n')

	norm1=norm1_in.copy()
	norm2=norm2_in.copy().T
	slip1=slip1_in.copy()
	slip2=slip2_in.copy().T

	num_vect=norm2.shape[0]
	rotemp=np.zeros((num_vect,4))
	for iter_x in range(0,4): # Iteration over the 4 possibilities
		if iter_x<2:
			norm2_temp=norm2.copy()
			slip2_temp=slip2.copy()
		else:
			norm2_temp=slip2.copy()
			slip2_temp=norm2.copy()
		if (iter_x==1) | (iter_x==3):
			norm2_temp=-norm2_temp
			slip2_temp=-slip2_temp

		B1=np.cross(slip1,norm1)*-1
		B2=np.cross(slip2_temp,norm2_temp)*-1

		phi=np.zeros((num_vect,3))
		phi[:,0]=norm1[0]*norm2_temp[:,0]+norm1[1]*norm2_temp[:,1]+norm1[2]*norm2_temp[:,2]
		phi[:,1]=slip1[0]*slip2_temp[:,0]+slip1[1]*slip2_temp[:,1]+slip1[2]*slip2_temp[:,2]
		phi[:,2]=B1[0]*B2[:,0]+B1[1]*B2[:,1]+B1[2]*B2[:,2]
		phi[phi>1]=1
		phi[phi<-1]=-1
		phi=np.arccos(phi)

		phi_flag=(phi<(1e-3))
		# if the mechanisms are very close, rotation = 0. Otherwise, calculate the rotation
		rot_ind=np.where(np.any(~phi_flag,axis=1))[0]

		# if one vector is the same, it is the rotation axis
		tmp_ind=rot_ind[np.where(phi_flag[rot_ind,2])[0]]
		rotemp[tmp_ind,iter_x]=(phi[tmp_ind,0])
		tmp_ind=rot_ind[np.where(phi_flag[rot_ind,0])[0]]
		rotemp[tmp_ind,iter_x]=(phi[tmp_ind,1])
		tmp_ind=rot_ind[np.where(phi_flag[rot_ind,1])[0]]
		rotemp[tmp_ind,iter_x]=(phi[tmp_ind,2])

		# find difference vectors - the rotation axis must be orthogonal to all three vectors
		rot_ind=np.where(np.all(~phi_flag,axis=1))[0]

		if len(rot_ind)==0:
			continue

		n=np.zeros((len(rot_ind),3,3))
		n[:,:,0]=norm1-norm2_temp[rot_ind,:]
		n[:,:,1]=slip1-slip2_temp[rot_ind,:]
		n[:,:,2]=B1-B2[rot_ind,:]
		scale=np.sqrt(n[:,0,:]**2+n[:,1,:]**2+n[:,2,:]**2)
		n=n/scale[:,np.newaxis,:]

		qdot=np.zeros((len(rot_ind),3))
		qdot[:,2]=n[:,0,0]*n[:,0,1]+n[:,1,0]*n[:,1,1]+n[:,2,0]*n[:,2,1]
		qdot[:,1]=n[:,0,0]*n[:,0,2]+n[:,1,0]*n[:,1,2]+n[:,2,0]*n[:,2,2]
		qdot[:,0]=n[:,0,1]*n[:,0,2]+n[:,1,1]*n[:,1,2]+n[:,2,1]*n[:,2,2]

		# use the two largest difference vectors, as long as they aren't orthogonal
		iout=np.zeros(len(rot_ind),dtype=int)-1
		qdot_flag=np.any(qdot>0.9999,axis=1)
		tmp_row=np.where(qdot_flag)[0]
		if len(tmp_row)>0:
			iout[tmp_row]=np.argmax(qdot[tmp_row,:],axis=1)
		tmp_row=np.where(~qdot_flag)[0]
		if len(tmp_row)>0:
			iout[tmp_row]=np.argmin(scale[tmp_row,:],axis=1)

		n1=np.zeros((len(rot_ind),3))
		n2=np.zeros((len(rot_ind),3))
		k=np.ones(len(rot_ind),dtype=bool)
		for j in range(3):
			tmp_ind=np.where(j!=iout)[0]
			tmp_ind_1=tmp_ind[k[tmp_ind]==True]
			tmp_ind_2=tmp_ind[k[tmp_ind]==False]

			if len(tmp_ind_1)>0:
				n1[tmp_ind_1,:]=n[tmp_ind_1,:,j]
				k[tmp_ind_1]=False
			if len(tmp_ind_2)>0:
				n2[tmp_ind_2,:]=n[tmp_ind_2,:,j]

		#  find rotation axis by taking cross product
		R=np.cross(n2,n1)*-1
		scaleR=np.sqrt(np.sum(R**2,axis=1))

		if np.any(scaleR==0):
			tmp_ind=np.where(scaleR==0)[0]
			rotemp[rot_ind[tmp_ind],iter_x]=9999
			rot_ind=np.delete(rot_ind,tmp_ind)
			scaleR=np.delete(scaleR,tmp_ind)
			R=np.delete(R,tmp_ind,axis=0)

		R=R/scaleR[:,np.newaxis]
		theta=np.zeros((len(rot_ind),3))
		theta[:,0]=norm1[0]*R[:,0]+norm1[1]*R[:,1]+norm1[2]*R[:,2]
		theta[:,1]=slip1[0]*R[:,0]+slip1[1]*R[:,1]+slip1[2]*R[:,2]
		theta[:,2]=B1[0]*R[:,0]+B1[1]*R[:,1]+B1[2]*R[:,2]
		theta[theta>1]=1
		theta[theta<-1]=-1
		theta=np.arccos(theta)

		iuse=np.argmin(np.abs(theta-(np.pi/2)),axis=1)
		tmp_ind=np.arange(len(iuse))

		tmp_rotemp=(np.cos(phi[rot_ind,iuse])-np.cos(theta[tmp_ind,iuse])**2)/(np.sin(theta[tmp_ind,iuse])**2)
		tmp_rotemp[tmp_rotemp>1]=1
		tmp_rotemp[tmp_rotemp<-1]=-1
		tmp_rotemp=np.arccos(tmp_rotemp)
		rotemp[rot_ind,iter_x]=tmp_rotemp

	rotemp=np.rad2deg(rotemp)
	rotemp=np.abs(rotemp)
	irot=np.argmin(rotemp,axis=1)

	tmp_ind=np.arange(len(irot))
	rota=rotemp[tmp_ind,irot]

	tmp_ind=np.where(irot>=2)[0]
	qtemp=slip2[tmp_ind,:]
	slip2[tmp_ind,:]=norm2[tmp_ind,:]
	norm2[tmp_ind,:]=qtemp

	tmp_ind=np.where( (irot==1) | (irot==3) )[0]
	norm2[tmp_ind,:]*=-1
	slip2[tmp_ind,:]*=-1

	return rota,norm2.T,slip2.T


def mech_probability(norm1in,norm2in,cangle,prob_max,iterative_avg=False):
	'''
	Determines the probability of mechanism solutions.

	Inputs:
		norm1in: normal to fault plane
		norm2in: slip vector
		cangle: cutoff angle
		prob_max: cutoff percent for mechanism multiples
		iterative_avg:
			if True: Compute average by removing one mechanism far from the average at a time following HASH
			if False: Compute average by considering the solutions within $cangle of the average.
	Outputs:
		mech_df: DataFrame of mechanism probabilities
	'''
	if norm1in.shape != norm2in.shape:
		raise ValueError('***Error in mech_probability: shape of norm1in and norm2in must be the same')
	if len(norm1in.shape)!=2:
		raise ValueError('***Error in mech_probability: norm1in and norm2in must each be an array of shape 3-by-n')
	if norm1in.shape[0]!=3:
		raise ValueError('***Error in mech_probability: norm1in and norm2in must each be an array of shape 3-by-n')

	nf=norm1in.shape[1]
	if nf==0:
		print('*mech_probability: norm1 is empty')
		return -1

	str_avg=np.zeros(5)-999
	dip_avg=np.zeros(5)-999
	rak_avg=np.zeros(5)-999
	prob=np.zeros(5)-999
	rms_diff=np.zeros((2,5))

	if nf==1: # If there's only one mech, return that mech
		str_avg[0],dip_avg[0],rak_avg[0]=sdr_from_vector(norm1in[:,[0]],norm2in[:,[0]])
		prob[0]=1
		mech_df=pd.DataFrame(columns=['str_avg','dip_avg','rak_avg','prob','rms_diff','rms_diff_aux'],index=[0])
		mech_df['str_avg']=str_avg[0]
		mech_df['dip_avg']=dip_avg[0]
		mech_df['rak_avg']=rak_avg[0]
		mech_df['prob']=prob[0]
		mech_df['rms_diff']=rms_diff[0][0]
		mech_df['rms_diff_aux']=rms_diff[1][0]
		return mech_df

	norm1=norm1in.copy()
	norm2=norm2in.copy()
	norm_ind=np.arange(norm1.shape[1])
	rota=np.zeros(nf)
	nsltn=0

	for imult in range(5):
		unused_norm_ind=[]
		if iterative_avg:
			for icount in range(len(norm_ind)):
				norm1_avg,norm2_avg=average_mech(norm1[:,norm_ind],norm2[:,norm_ind])
				temp_rota,temp1,temp2=mech_rotation(norm1_avg,norm1[:,norm_ind],norm2_avg,norm2[:,norm_ind])

				temp_rota=np.abs(temp_rota)

				imax=np.argmax(temp_rota)
				imax_ind=norm_ind[imax]
				maxrot=temp_rota[imax]

				if maxrot<=cangle:
					break
				else:
					unused_norm_ind.append(imax_ind)
					norm_ind=np.delete(norm_ind,imax)
			prob[imult]=len(norm_ind)/nf
		else:
			# Compute the average mech of the solutions and then determine the rotation angle for each solution
			norm1_avg,norm2_avg=average_mech(norm1[:,norm_ind],norm2[:,norm_ind])
			temp_rota,temp1,temp2=mech_rotation(norm1_avg,norm1[:,norm_ind],norm2_avg,norm2[:,norm_ind])
			tmp_ind=norm_ind[np.abs(temp_rota)<=cangle]

			if len(tmp_ind)==0:
				break

			# Considering the solutions within $cangle of the average mech, recompute the average mech and find similar solutions
			norm1_avg,norm2_avg=average_mech(norm1[:,tmp_ind],norm2[:,tmp_ind])
			temp_rota,temp1,temp2=mech_rotation(norm1_avg,norm1[:,norm_ind],norm2_avg,norm2[:,norm_ind])
			tmp2_flag=(np.abs(temp_rota)<=cangle)
			norm1_avg,norm2_avg=average_mech(norm1[:,norm_ind[tmp2_flag]],norm2[:,norm_ind[tmp2_flag]])

			unused_norm_ind=norm_ind[~tmp2_flag]
			norm_ind=norm_ind[tmp2_flag]

			prob[imult]=np.sum(tmp2_flag)/nf

		if (imult>0) & (prob[imult]<prob_max):
			break

		# determine the RMS observed angular difference between the average
		# normal vectors and the normal vectors of each mechanism
		rota,temp1,temp2=mech_rotation(norm1_avg,norm1in,norm2_avg,norm2in)
		d11=temp1[0]*norm1_avg[0]+temp1[1]*norm1_avg[1]+temp1[2]*norm1_avg[2]
		d22=temp2[0]*norm2_avg[0]+temp2[1]*norm2_avg[1]+temp2[2]*norm2_avg[2]

		d11[d11>1]=1
		d11[d11<-1]=-1
		d22[d22>1]=1
		d22[d22<-1]=-1

		a11=np.arccos(d11)
		a22=np.arccos(d22)
		rms_diff[0,imult]=np.rad2deg(np.sqrt(np.sum(a11**2)/nf))
		rms_diff[1,imult]=np.rad2deg(np.sqrt(np.sum(a22**2)/nf))

		str_avg[imult],dip_avg[imult],rak_avg[imult]=sdr_from_vector(norm1_avg[np.newaxis].T,norm2_avg[np.newaxis].T)
		nsltn+=1

		if len(unused_norm_ind)>0:
			norm_ind=np.asarray(unused_norm_ind)
		else:
			break
	str_avg=str_avg[:nsltn]
	dip_avg=dip_avg[:nsltn]
	rak_avg=rak_avg[:nsltn]
	prob=prob[:nsltn]
	rms_diff=rms_diff[:,:nsltn]

	sort_ind=np.argsort(prob)[::-1]

	# Returns result as a DataFrame
	mech_df=pd.DataFrame(columns=['str_avg','dip_avg','rak_avg','prob','rms_diff','rms_diff_aux'],index=np.arange(len(sort_ind)))
	mech_df['str_avg']=str_avg[sort_ind]
	mech_df['dip_avg']=dip_avg[sort_ind]
	mech_df['rak_avg']=rak_avg[sort_ind]
	mech_df['prob']=prob[sort_ind]
	mech_df['rms_diff']=rms_diff[0][sort_ind]
	mech_df['rms_diff_aux']=rms_diff[1][sort_ind]

	return mech_df


def focal_gridsearch(sr_azimuth,takeoff,p_pol,sp_amp,dir_cos_dict,nextra,ntotal,qextra,qtotal,maxout,ncoor,min_ratio_trial_solutions=0.5,min_num_sp_solutions=10):
	'''
	Performs a grid search to find focal mechanisms using P-polarity and S/P ratio information using the python routine.
	Input:
		sr_azimuth: source-receiver azimuths, 2d array
		takeoff: takeoff angles, 2d array
		p_pol: polarity weights, 1d array
		sp_amp: polarity weights, 1d array
		dir_cos_dict: coordinate transformation dictionary, produced by dir_cos_setup()
		nextra: number of polarity additional misfits allowed above minimum
		ntotal: total number of allowed polarity misfits
		qextra: additional amplitude misfit allowed above minimum
		qtotal: total allowed amplitude misfit
		maxout: maximum number of fault planes to return
		ncoor: number of test mechanisms
		min_ratio_trial_solutions: minimum ratio of trial solutions from polarities before criteria loosened
		min_num_sp_solutions: minimum ratio of trial solutions from S/P ratios before criteria loosened
	Output:
		faultnorms_all: fault normal vectors
		faultslips_all: fault slip vectors
	'''
	ntab=180
	astep=1/ntab

	'''
	Transforms the spherical coordinates to cartesian
	'''
	pol_ind=np.where(p_pol!=0)[0]

	takeoff_r=np.deg2rad(takeoff[pol_ind])
	sr_azimuth_r=np.deg2rad(sr_azimuth[pol_ind])

	xyz=np.zeros((3,takeoff_r.shape[0],takeoff_r.shape[1]))
	xyz[0,:]=np.sin(takeoff_r)*np.cos(sr_azimuth_r)
	xyz[1,:]=np.sin(takeoff_r)*np.sin(sr_azimuth_r)
	xyz[2,:]=-np.cos(takeoff_r)

	p_b1=np.tensordot(xyz,dir_cos_dict['b1'],axes=[[0],[0]])
	p_b3=np.tensordot(xyz,dir_cos_dict['b3'],axes=[[0],[0]])

	# # Slower (but probably more comprehensible) implementation of the above
	# p_b1=b1[0,:]*xyz[0,:,:,np.newaxis]+\
	#         b1[1,:]*xyz[1,:,:,np.newaxis]+\
	#         b1[2,:]*xyz[2,:,:,np.newaxis]
	# p_b3=b3[0,:]*xyz[0,:,:,np.newaxis]+\
	#         b3[1,:]*xyz[1,:,:,np.newaxis]+\
	#         b3[2,:]*xyz[2,:,:,np.newaxis]

	prod=((p_b1<0)!=(p_b3<0)) # If True, predicted sign is negative. If False, predicted sign is positive

	qmiss=(prod != (p_pol[pol_ind]<0)[:,np.newaxis,np.newaxis])*np.abs(p_pol[pol_ind])[:,np.newaxis,np.newaxis]

	fit=np.sum(qmiss,axis=0)

	# Calculates max misfit for each trial
	qmissmax=fit.min(axis=1)+nextra
	qmissmax[qmissmax<ntotal]=ntotal

	sp_finite_ind=np.where(np.isfinite(sp_amp))[0]
	qacount=len(sp_finite_ind)
	if qacount>0:
		takeoff_r=np.deg2rad(takeoff[sp_finite_ind])
		sr_azimuth_r=np.deg2rad(sr_azimuth[sp_finite_ind])

		xyz=np.zeros((3,takeoff_r.shape[0],takeoff_r.shape[1]))
		xyz[0,:]=np.sin(takeoff_r)*np.cos(sr_azimuth_r)
		xyz[1,:]=np.sin(takeoff_r)*np.sin(sr_azimuth_r)
		xyz[2,:]=-np.cos(takeoff_r)

		p_b3=np.tensordot(xyz,dir_cos_dict['b3'],axes=[[0],[0]])

		p_proj1=xyz[0,:,:,np.newaxis]-(p_b3*dir_cos_dict['b3'][0,:])
		p_proj2=xyz[1,:,:,np.newaxis]-(p_b3*dir_cos_dict['b3'][1,:])
		p_proj3=xyz[2,:,:,np.newaxis]-(p_b3*dir_cos_dict['b3'][2,:])

		plen=np.sqrt(p_proj1**2+p_proj2**2+p_proj3**2)
		p_proj1=p_proj1/plen
		p_proj2=p_proj2/plen
		p_proj3=p_proj3/plen

		pp_b1=dir_cos_dict['b1'][0,:]*p_proj1+dir_cos_dict['b1'][1,:]*p_proj2+dir_cos_dict['b1'][2,:]*p_proj3
		pp_b2=dir_cos_dict['b2'][0,:]*p_proj1+dir_cos_dict['b2'][1,:]*p_proj2+dir_cos_dict['b2'][2,:]*p_proj3
		i=np.round((p_b3+1.)/astep).astype(int)

		theta=dir_cos_dict['thetable'][i]
		i=np.round((pp_b2+1.)/astep).astype(int)
		j=np.round((pp_b1+1.)/astep).astype(int)
		phi=dir_cos_dict['phitable'][i,j]

		i=np.round(phi/(np.pi*astep)).astype(int)
		i[i>(2*ntab-1)]=0
		j=np.round(theta/(np.pi*astep)).astype(int)
		j[j>(ntab-1)]=0

		p_amp=dir_cos_dict['amptable'][0,j,i]
		s_amp=dir_cos_dict['amptable'][1,j,i]

		sp_rat=np.zeros(p_amp.shape)
		sp_rat[p_amp==0]=4.0
		sp_rat[s_amp==0]=-2.0
		nonzero_flag=((p_amp!=0) & (s_amp!=0))
		sp_rat[nonzero_flag]=np.log10(4.9*s_amp[nonzero_flag]/p_amp[nonzero_flag])

		qamiss=np.abs(sp_amp[sp_finite_ind][:,np.newaxis,np.newaxis]-sp_rat)
		afit=np.sum(qamiss,axis=0)

		# Calculates max misfit for each trial
		qamissmax=afit.min(axis=1)+qextra
		qamissmax[qamissmax<qtotal]=qtotal

		goodmech_flag=( (fit<=qmissmax[:,np.newaxis]) & (afit<=qamissmax[:,np.newaxis]) )

		# Ratio of the number of trials with a solution to trials with no solution
		ratio_trials_with_solutions=np.sum(np.any(goodmech_flag,axis=1))/goodmech_flag.shape[0]
		# Number of unique mech solutions across the trials
		good_fp_ind=np.where(np.any(goodmech_flag,axis=0))[0]

		# If there are no solutions that meet the criteria, loosen the amplitude criteria
		if (len(good_fp_ind)<min_num_sp_solutions) | (ratio_trials_with_solutions<min_ratio_trial_solutions):
			masked_afit=np.ma.masked_array(afit,mask=(fit>qmissmax[:,np.newaxis]))
			qmis0min=masked_afit.min(axis=1).data
			qamissmax=qmis0min+nextra
			qamissmax[qamissmax<qtotal]=qtotal
			goodmech_flag=( (fit<=qmissmax[:,np.newaxis]) & (afit<=qamissmax[:,np.newaxis]) )
			good_fp_ind=np.where(np.any(goodmech_flag,axis=0))[0]

	else:
		goodmech_flag=( (fit<=qmissmax[:,np.newaxis]) )
		good_fp_ind=np.where(np.any(goodmech_flag,axis=0))[0]

	if len(good_fp_ind)>maxout: # If more than maxout solutions meet criteria, randomly select maxout solutions
		good_fp_ind=rng.choice(good_fp_ind,maxout,replace=False)

	faultnorms_all=np.vstack((dir_cos_dict['b3'][0,good_fp_ind],dir_cos_dict['b3'][1,good_fp_ind],dir_cos_dict['b3'][2,good_fp_ind]))
	faultslips_all=np.vstack((dir_cos_dict['b1'][0,good_fp_ind],dir_cos_dict['b1'][1,good_fp_ind],dir_cos_dict['b1'][2,good_fp_ind]))

	return faultnorms_all,faultslips_all


def determine_max_gap(in_azimuth_deg,in_takeoff_deg):
	'''
	Given an array of azimuths and takeoff angles, calculates the maximum
	azimuthal and takeoff gaps.
	Input:
		in_azimuth_deg: source-receiver azimuths
		in_takeoff_deg: takeoff angles
	Output:
		max_azimuthal_gap: maximum azimuthal gap between measurements
		max_takeoff_gap: maximum takeoff angle gap between measurements
	'''
	azimuth_deg=in_azimuth_deg.copy()
	takeoff_deg=in_takeoff_deg.copy()
	azimuth_deg[takeoff_deg>90]-=180
	takeoff_deg[takeoff_deg>90]=180-takeoff_deg[takeoff_deg>90]
	azimuth_deg[azimuth_deg<0]+=360

	azimuth_deg=np.sort(azimuth_deg)
	takeoff_deg=np.sort(takeoff_deg)

	max_azimuthal_gap=np.max([np.max(np.diff(azimuth_deg)),azimuth_deg[0]+360-azimuth_deg[-1]])
	max_takeoff_gap=np.max([np.max(np.diff(takeoff_deg)),takeoff_deg[0],90-takeoff_deg[-1]])

	return max_azimuthal_gap,max_takeoff_gap


def mech_misfit(mech_df,p_azi_mc,p_the_mc,p_pol,sp_amp):
	'''
	Loops over the potential mech solution dataframe and calcultes the polarity
	and S/P misfits. The polarity and S/P agreements for the first mech solutions
	are returned.
	Input:
		mech_df: mechanism solution dataframe, produced by mech_probability()
		p_azi_mc: source-receiver azimuth for the measurement, 1d array
		p_the_mc: takeoff angles for the measurement, 1d array
		p_pol: polarity weights, 1d array
		sp_amp: polarity weights, 1d array
	Output:
		mech_df: mechanism solution dataframe, with additional columns
		pol_agreement_out: Flag for if the polarity measurement (dis)agrees with the solution
		sp_diff_out: Flag for if the S/P measurement (dis)agrees with the solution
	'''
	mech_df['mfrac']=np.nan
	mech_df['mavg']=np.nan
	mech_df['stdr']=np.nan

	for imult in range(len(mech_df)):
		mech_df.loc[imult,'mfrac'],mech_df.loc[imult,'mavg'],mech_df.loc[imult,'stdr'],tmp_pol_agreement_out,tmp_sp_diff_out=calculate_misfit(p_azi_mc,p_the_mc,p_pol,sp_amp,mech_df.loc[imult,'str_avg'],mech_df.loc[imult,'dip_avg'],mech_df.loc[imult,'rak_avg'])
		if imult==0:
			pol_agreement_out=tmp_pol_agreement_out
			sp_diff_out=tmp_sp_diff_out
	return mech_df,pol_agreement_out,sp_diff_out


def mech_quality(mech_df,qual_criteria_dict):
	'''
	Given the mech solutions and the quality criteria, adds the mech qualities
	to the mech solution dataframe.
	Input:
		mech_df: mechanism solution dataframe, produced by mech_probability()
		qual_criteria_dict: dictionary of quality criteria, created in SKHASH.py
	Output:
		mech_df: mechanism solution dataframe, with additional quality column
	'''
	qual_ind=np.zeros((len(mech_df),4),dtype=int)-999
	qual_ind[:,0]=np.argmax(mech_df.prob.values[:,np.newaxis]>=qual_criteria_dict['probs'],axis=1)
	qual_ind[:,1]=np.argmax(mech_df[['rms_diff','rms_diff_aux']].mean(axis=1).values[:,np.newaxis]<=qual_criteria_dict['var_avg'],axis=1)
	qual_ind[:,2]=np.argmax(mech_df.mfrac.values[:,np.newaxis]<=qual_criteria_dict['mfrac'],axis=1)
	qual_ind[:,3]=np.argmax(mech_df.stdr.values[:,np.newaxis]>=qual_criteria_dict['stdr'],axis=1)
	mech_df['qual']=qual_criteria_dict['qual_letter'][np.max(qual_ind,axis=1)]
	# If the following conditions are not met, the lowest quality letter is given.
	mech_df.loc[((mech_df['azimuthal_gap']>qual_criteria_dict['azimuthal_gap']) |
			 	(mech_df['takeoff_gap']>qual_criteria_dict['takeoff_gap']) |
				(mech_df['num_p_pol']<qual_criteria_dict['num_p_pol'])
				),'qual']=qual_criteria_dict['qual_letter'][-1]

	return mech_df


def calculate_misfit(p_azi_mc,p_the_mc,p_pol,sp_amp,str_avg,dip_avg,rak_avg):
	'''
	Calculates the polarity misfit percent and S/P misfit for a given mechanism solution.

	Inputs:
		p_azi_mc: azimuths
		p_the_mc: takeoff angles
		p_pol: polarity measurements
		sp_amp: S/P ratios
		str_avg,dip_avg,rak_avg: the given mechanism solution

	Outputs:
		mfrac: weighted fraction misfit polarities
		mavg: average S/P misfit (log10)
		stdr: station distribution ratio
		pol_agreement_out: boolean of whether the polarity agrees with the solution
		sp_diff_out: difference between the measured and expected S/P ratio given the solution
	'''

	strike_r=np.deg2rad(str_avg)
	dip_r=np.deg2rad(dip_avg)
	rake_r=np.deg2rad(rak_avg)

	M=np.zeros((3,3))
	M[0,0]=-np.sin(dip_r)*np.cos(rake_r)*np.sin(2*strike_r)-np.sin(2*dip_r)*np.sin(rake_r)*np.sin(strike_r)*np.sin(strike_r)
	M[1,1]=np.sin(dip_r)*np.cos(rake_r)*np.sin(2*strike_r)-np.sin(2*dip_r)*np.sin(rake_r)*np.cos(strike_r)*np.cos(strike_r)
	M[2,2]=np.sin(2*dip_r)*np.sin(rake_r)
	M[0,1]=np.sin(dip_r)*np.cos(rake_r)*np.cos(2*strike_r)+0.5*np.sin(2*dip_r)*np.sin(rake_r)*np.sin(2*strike_r)
	M[1,0]=M[0,1]
	M[0,2]=-np.cos(dip_r)*np.cos(rake_r)*np.cos(strike_r)-np.cos(2*dip_r)*np.sin(rake_r)*np.sin(strike_r)
	M[2,0]=M[0,2]
	M[1,2]=-np.cos(dip_r)*np.cos(rake_r)*np.sin(strike_r)+np.cos(2*dip_r)*np.sin(rake_r)*np.cos(strike_r)
	M[2,1]=M[1,2]

	bb3,bb1=vector_from_sdr(strike_r,dip_r,rake_r)
	bb2=np.cross(bb1,bb3)*-1

	mfrac=0.
	qcount=0.
	stdr=0.
	scount=0.
	mavg=0.
	acount=0.

	p_a1,p_a2,p_a3=cartesian_transform(p_the_mc,p_azi_mc,1)
	p_b1=bb1[0]*p_a1+bb1[1]*p_a2+bb1[2]*p_a3
	p_b3=bb3[0]*p_a1+bb3[1]*p_a2+bb3[2]*p_a3
	p_proj1=p_a1-p_b3*bb3[0]
	p_proj2=p_a2-p_b3*bb3[1]
	p_proj3=p_a3-p_b3*bb3[2]
	plen=np.sqrt(p_proj1*p_proj1+p_proj2*p_proj2+p_proj3*p_proj3)
	p_proj1=p_proj1/plen
	p_proj2=p_proj2/plen
	p_proj3=p_proj3/plen
	pp_b1=bb1[0]*p_proj1+bb1[1]*p_proj2+bb1[2]*p_proj3
	pp_b2=bb2[0]*p_proj1+bb2[1]*p_proj2+bb2[2]*p_proj3
	phi=np.arctan2(pp_b2,pp_b1)
	theta=np.arccos(p_b3)
	p_amp=np.abs(np.sin(2*theta)*np.cos(phi))

	pol_ind=np.where(p_pol!=0)[0]
	wt=np.sqrt(p_amp)

	scount=np.sum(np.abs(p_pol[pol_ind]))
	a=np.zeros((3,len(pol_ind)))
	b=np.zeros((3,len(pol_ind)))

	azi=np.deg2rad(p_azi_mc[pol_ind])
	toff=np.deg2rad(p_the_mc[pol_ind])
	a[0,:]=np.sin(toff)*np.cos(azi)
	a[1,:]=np.sin(toff)*np.sin(azi)
	a[2,:]=-np.cos(toff)
	b=np.sum(M[:,:,np.newaxis]*a,axis=1)

	neg_pol_v=(a[0,:]*b[0,:]+a[1,:]*b[1,:]+a[2,:]*b[2,:])*p_pol[pol_ind]
	mfrac=np.sum(wt[pol_ind][neg_pol_v<0])

	pol_agreement_out=np.empty(len(p_pol))
	pol_agreement_out[:]=np.nan
	pol_agreement_out[pol_ind[neg_pol_v>=0]]=True
	pol_agreement_out[pol_ind[neg_pol_v<0]]=False

	qcount=np.sum(wt[pol_ind])
	stdr=np.sum(wt[pol_ind]*np.abs(p_pol[pol_ind]))

	sp_ind=np.where(~np.isnan(sp_amp))[0]
	acount=len(sp_ind)
	sp_diff_out=np.empty(len(sp_amp),dtype=float)
	sp_diff_out[:]=np.nan
	if acount>0:
		s1=np.cos(2*theta[sp_ind])*np.cos(phi[sp_ind])
		s2=-np.cos(theta[sp_ind])*np.sin(phi[sp_ind])
		s_amp=np.sqrt(s1*s1+s2*s2)
		sp_rat=np.log10(4.9*s_amp/p_amp[sp_ind])

		#Difference between S/P ratio and expected given the solution
		sp_diff=sp_amp[sp_ind]-sp_rat
		sp_diff_out[sp_ind]=sp_diff

		mavg=np.sum(sp_diff)
		stdr+=np.sum(wt[sp_ind])
		scount+=acount

	if qcount==0:
		mfrac=0
	else:
		mfrac=mfrac/qcount

	if acount==0:
		mavg=0
	else:
		mavg=mavg/acount

	if scount==0:
		stdr=0
	else:
		stdr=stdr/scount

	return mfrac,mavg,stdr,pol_agreement_out,sp_diff_out


def create_lookup_table(p_dict):
	'''
	Reads in the velocity model inputs, does minor QCing, and creates lookup table (if necessary).
	If the lookup table already has been computed, it can read it from the disk.

	Inputs:
		p_dict['vmodel_paths']: a list of filepaths to the velocity models. Each file should be a whitespace delimited file following the format:
			depth(km) vp(km/s)
		p_dict['look_dep']: list of: minimum source depth, maximum, and interval for the lookup table
		p_dict['look_del']: list of: minimum source-station distance, maximum, and interval for the lookup table
		p_dict['nump']: number of rays traced
		p_dict['nx0']: maximum source-station distance (km) bins for look-up tables
		p_dict['nd0']: maximum source depth (km) bins for look-up tables
		p_dict['output_angle_precision']: Number of decimal places to output for take off angles
		p_dict['recompute_lookup_table']: a boolean. If True, the lookup tables will be recomputed, even if a matching lookup table exists.
		p_dict['write_lookup_table']: a boolean. If True, the lookup table will be written to the disk. A '.npz' file suffix will be added to the path.
	Output:
		deptab: array of depths (km)
		delttab: array of distances (km)
		table: Lookup table (array)
	'''

	deptab=np.arange(p_dict['look_dep'][0],p_dict['look_dep'][1]+p_dict['look_dep'][2],p_dict['look_dep'][2]) # array of depths (km)
	delttab=np.arange(p_dict['look_del'][0],p_dict['look_del'][1]+p_dict['look_del'][2],p_dict['look_del'][2]) # array of distances (km)

	if (p_dict['recompute_lookup_table']==False) or (p_dict['write_lookup_table']):
		lookup_table_params=np.asarray([p_dict['nump'],p_dict['nx0'],p_dict['nd0'],p_dict['output_angle_precision']])

	table_list=[]
	for vmodel_ind,vmodel_path in enumerate(p_dict['vmodel_paths']):
		lookup_vmodel_path=vmodel_path+'.lookup.npz'
		vmodel_depthvp=pd.read_csv(vmodel_path,names=['depth','vp_km_s'],sep=',',comment='#').values
		vmodel_depthvp[:,0]=np.round(vmodel_depthvp[:,0],5) # Rounds velocity model depths
		if len(vmodel_depthvp)==1:
			raise ValueError('***Velocity model ({}) has only a single velocity. There must be at least two points.***'.format(vmodel_path))
		else:
			if not(all(np.diff(vmodel_depthvp[:,0])>=0)):
				print('*WARNING: The velocity model ({}) is expected to be ordered in terms of increasing depth.'.format(vmodel_path))
				vm_sort_ind=np.argsort(vmodel_depthvp[:,0])
				vmodel_depthvp=vmodel_depthvp[vm_sort_ind,:]
			if np.any(np.diff(vmodel_depthvp[:,0])==0):
				print('*WARNING: The velocity model ({}) has duplicate velocities for a given depth.'.format(vmodel_path))
				vmodel_depthvp=np.delete(vmodel_depthvp,np.where(np.diff(vmodel_depthvp[:,0])==0)[0],axis=0)
		if vmodel_depthvp[:,0].max()>6400:
			print('*WARNING: The velocity model ({}) has depths as large as {} km. Is this intentional?'.format(vmodel_path,vmodel_depthvp[:,0].max()))

		# If there are constant velocity layers, merge those rows
		drop_constant_vel_ind=np.where(np.diff(vmodel_depthvp[:,1])<=0)[0]+1
		if len(drop_constant_vel_ind)>0:
			vmodel_depthvp=np.delete(vmodel_depthvp,drop_constant_vel_ind,axis=0)

		compute_lookup_table=True
		if os.path.exists(lookup_vmodel_path) & (p_dict['recompute_lookup_table']==False):
			print('Loading precomputed lookup table ({}/{}): {}'.format(vmodel_ind,len(p_dict['vmodel_paths'])-1,lookup_vmodel_path))
			try:
				tmp_lookup=np.load(lookup_vmodel_path)
			except:
				print('\tError loading lookup table. Recomputing.')

			# Ensures the lookup table .npz file contains the expected variable names
			if tmp_lookup.files==['table', 'lookup_table_params', deptab, delttab]:
				# Ensures that the parameters used to create the lookup table are identical to the current parameters.
				if np.array_equal(tmp_lookup['lookup_table_params'],lookup_table_params) &\
					np.array_equal(tmp_lookup[deptab],deptab) &\
					np.array_equal(tmp_lookup[delttab],delttab):
					table=tmp_lookup['table']
					compute_lookup_table=False
					print('\tLoad successful.')
				else:
					print('\tThe parameters used to create the lookup table differ from your current set parameters. Recomputing lookup table.')
			else:
				print('\tIssue with interpreting previously saved lookup table. Recomputing lookup table.')

		if compute_lookup_table:
			print('Creating lookup table ({}/{}): {}'.format(vmodel_ind,len(p_dict['vmodel_paths'])-1,vmodel_path))
			table=create_takeoff_table(vmodel_depthvp,deptab,delttab,p_dict['nump'],p_dict['nx0'],p_dict['nd0'],p_dict['output_angle_precision'])
			print('\tCreated table.')

			if p_dict['write_lookup_table']:
				np.savez(lookup_vmodel_path,table=table,lookup_table_params=lookup_table_params,deptab=deptab,delttab=delttab)
				print('\tSaved lookup table: {}'.format(lookup_vmodel_path))
		table_list.append(table)
	table=np.dstack(table_list)

	return {'deptab':deptab,'delttab':delttab,'table':table}



'''
Functions for reading station files and station reversals.
'''

# Standard libraries
import time

# External libraries
import numpy as np
import pandas as pd


def read_station_file(pol_df,p_dict):
	'''
	Reads input station file.
	'''
	if (p_dict['input_format_stfile']=='skhash'):
		station_df=read_skhash_station_file(p_dict['stfile'],p_dict['merge_on'])
	elif (p_dict['input_format_stfile']=='hypoinverse'):
		station_df=read_hypoinverse_station_file(p_dict['stfile'])
	elif (p_dict['input_format_stfile']=='hash2') | (p_dict['input_format_stfile']=='hash3'):
		station_df=read_hash3_station_file(p_dict['stfile'])
	elif (p_dict['input_format_stfile']=='hash4'):
		station_df=read_hash4_station_file(p_dict['stfile'])
	else:
		raise ValueError('Unknown station metadata file format ({}).'.format(p_dict['input_format_stfile']))

	# Ensures there is no whitespace in net/sta/loc/cha codes
	for merge_col in p_dict['merge_on']:
		station_df[merge_col]=station_df[merge_col].str.strip()

	station_df[p_dict['merge_on']]=station_df[p_dict['merge_on']].astype(str)
	station_df['sta_code']=station_df[p_dict['merge_on']].agg('.'.join, axis=1)

	# Ensures the required columns are present
	if not({'sta_code','station_lat','station_lon','station_depth_km'}).issubset(station_df.columns):
		raise ValueError('The station dataframe does not contain the required columns:\n\t[sta_code, station_lat, station_lon, station_depth_km]')

	if p_dict['require_temporal_match']:
		if not({'start_time','end_time'}).issubset(station_df.columns):
			raise ValueError('When require_temporal_match==True, the station dataframe must contain the additional columns:\n\t[start_time, end_time]')
		# Converts the start/end time strings into timestamps.
		station_df['start_time']=pd.to_datetime(station_df['start_time'],infer_datetime_format=True,errors='coerce')
		station_df['start_time']=station_df['start_time'].fillna(pd.Timestamp('1700-01-01'))
		station_df['end_time']=pd.to_datetime(station_df['end_time'],infer_datetime_format=True,errors='coerce')
		station_df['end_time']=station_df['end_time'].fillna(pd.Timestamp('2200-01-01'))

		# Selects only the desired columns
		station_df=station_df.filter(['sta_code','station_lat','station_lon','station_depth_km','start_time','end_time'])
	else:
		# Selects only the desired columns
		station_df=station_df.filter(['sta_code','station_lat','station_lon','station_depth_km'])

	# Selects only the station metadata that has polarity or S/P amp information
	station_df=station_df.loc[station_df['sta_code'].isin(pol_df['sta_code'].unique()),:].reset_index(drop=True)

	# Drops duplicate records
	station_df=station_df.drop_duplicates().reset_index(drop=True)

	return station_df


def read_skhash_station_file(stfile,merge_on):
	'''
	Reads input station file following the SKHASH format.
	'''
	station_df=pd.read_csv(stfile)
	required_column_names=['latitude','longitude','elevation']
	if not(set(required_column_names).issubset(station_df.columns)):
		raise ValueError('When using the SKHASH input format for station metadata files, the following header names MUST be provided:\n{}'.format(required_column_names))
	for req_col in merge_on:
		if not(req_col in station_df.columns):
			raise ValueError('When using SKHASH input format and require_{}_match==True, the station file (stfile: {}) must contain the column "{}".'.format(req_col,stfile,req_col))

	station_df=station_df.rename(columns={'latitude':'station_lat','longitude':'station_lon','elevation':'station_elevation_m'})
	station_df['station_depth_km']=station_df['station_elevation_m']/1000*-1 # Converts station elevation (in m) to station depth (in km)
	station_df=station_df.drop(columns='station_elevation_m')

	return station_df

def read_hypoinverse_station_file(stfile):
	'''
	Reads input station file following HYPOINVERSE station data format #2.

	If there are strings in the 82-85 columns, that text will be treated as
	an alternative channel name.
	'''
	station_df=pd.read_fwf(stfile,colspecs=[[0,5],[6,8],[10,13],[15,17],[18,25],[25,26],[26,29],[30,37],[37,38],[38,42],[80,82],[82,85]],
						names=['station','network','channel','station_lat_deg','station_lat_min','latNS','station_lon_deg','station_lon_min','lonEW','station_elevation_m','location','channel2'])
	station_df['latNS']=station_df['latNS'].fillna('N')
	station_df['lonEW']=station_df['lonEW'].fillna('W')
	station_df.loc[station_df['latNS']!='S','latNS']='N'
	station_df.loc[station_df['lonEW']!='E','lonEW']='W'
	station_df['station_lat']=dm2dd_vec(station_df['station_lat_deg'].values,station_df['station_lat_min'].values,station_df['latNS'].values)
	station_df['station_lon']=dm2dd_vec(station_df['station_lon_deg'].values,station_df['station_lon_min'].values,station_df['lonEW'].values)
	station_df['station_depth_km']=station_df['station_elevation_m']/1000*-1
	station_df['start_time']=pd.Timestamp('1900-01-01')
	station_df['end_time']=pd.Timestamp('2200-01-01')

	# Create new entries using channel2
	if any(~pd.isnull(station_df['channel2'])):
		tmp_ind=station_df[~pd.isnull(station_df['channel2'])].index
		tmp_ind=tmp_ind[station_df.loc[tmp_ind,'channel2'].str.len()>0]
		station2_df=station_df.loc[tmp_ind,:].copy()
		station2_df['channel']=station2_df['channel2']
		station_df=pd.concat([station_df,station2_df]).reset_index(drop=True)

	# # Replace normal channel with channel2
	# if any(~pd.isnull(station_df['channel2'])):
	# 	tmp_ind=station_df[~pd.isnull(station_df['channel2'])].index
	# 	tmp_ind=tmp_ind[station_df.loc[tmp_ind,'channel2'].str.len()>0]
	# 	station_df.loc[tmp_ind,'channel']=station_df.loc[tmp_ind,'channel2']

	station_df=station_df.loc[:,['station','network','location','channel','station_lat','station_lon','station_depth_km','start_time','end_time']]

	station_df=station_df.drop_duplicates().reset_index(drop=True)
	return station_df

def read_hash3_station_file(stfile):
	'''
	Reads input station file following the (old) SCEDC format used for HASH Driver 2, 3, and 5.
	'''
	station_df=pd.read_fwf(stfile,colspecs=[[0,4],[5,8],[41,50],[51,61],[62,67],[68,78],[79,89],[90,92]],names=['station','channel','station_lat','station_lon','station_elevation_m','start_time','end_time','network'])
	station_df['network']=station_df['network'].str.strip()
	station_df['station']=station_df['station'].str.strip()
	station_df['channel']=station_df['channel'].str.strip()
	station_df['station_depth_km']=station_df['station_elevation_m']/1000*-1
	station_df.drop(columns='station_elevation_m',inplace=True)
	station_df=station_df.loc[:,['station','network','channel','station_lat','station_lon','station_depth_km','start_time','end_time']]

	return station_df


def read_hash4_station_file(stfile):
	'''
	Reads input station file following the (newer) SCEDC format used for HASH Driver 4.
	'''
	station_df=pd.read_fwf(stfile,colspecs=[[0,2],[4,9],[10,13],[16,18],[51,60],[60,71],[71,77],[77,88],[88,99]],names=['network','station','channel','location','station_lat','station_lon','station_elevation_m','start_time','end_time'])
	station_df['network']=station_df['network'].str.strip()
	station_df['station']=station_df['station'].str.strip()
	station_df['channel']=station_df['channel'].str.strip()
	station_df['station_depth_km']=station_df['station_elevation_m']/1000*-1
	station_df.drop(columns='station_elevation_m',inplace=True)
	station_df=station_df.loc[:,['station','network','channel','station_lat','station_lon','station_depth_km','start_time','end_time']]

	return station_df


def apply_station_locations(pol_df,station_df,p_dict):
	'''
	Mergest the station locations to the polarity information.
	'''
	# Checks to see if there is missing metadata for the polarities
	missing_metadata_index=pol_df[~(pol_df['sta_code'].isin(station_df['sta_code']))].index
	if len(missing_metadata_index)>0:
		missing_sta_df=pol_df.loc[missing_metadata_index,'sta_code'].drop_duplicates().reset_index(drop=True)
		num_missing_sta=len(missing_sta_df)
		if p_dict['ignore_missing_metadata']:
			print('WARNING: Missing metadata for {} measurements. These measurements will be ignored. Example missing stations:\n{}'.format(num_missing_sta,missing_sta_df.head(10).to_string()))
			pol_df=pol_df.drop(missing_metadata_index).reset_index(drop=True)
		else:
			raise ValueError('Missing metadata for {} measurements. Example missing stations:\n{}'.format(num_missing_sta,missing_sta_df.head(10).to_string()))

	if p_dict['require_temporal_match']:
		time_merge_event_sta=time.time()

		pol_df=pol_df.reset_index(drop=True)
		station_df=station_df.reset_index(drop=True)

		# Merges measurement records and station metadata considering receiver start/end times.
		# This is slow and memory intensive, and it could be greatly improved!
		i, j = np.where((pol_df.origin_DateTime.values[:, None] >= station_df.start_time.values) &
						(pol_df.origin_DateTime.values[:, None] <= station_df.end_time.values) &
						(pol_df.sta_code.values[:,None]==station_df.sta_code.values))

		found_pol_df=pd.concat([pol_df.loc[i, :].reset_index(drop=True),
							station_df.loc[j, ['station_lat','station_lon','station_depth_km','start_time','end_time']].reset_index(drop=True)], axis=1)
		missing_pol_df=pol_df.iloc[~np.in1d(np.arange(len(pol_df)), np.unique(i)),:]

		if len(missing_pol_df)>0:
			print(('{} measurements did not match station metadata considering station start/end times.\n'+
					'Ignoring start/end times for these will be ignored, and the first matching metadata record will be used instead.'+
					'Example record:\n{}\n').format(len(missing_pol_df),missing_pol_df.loc[:,['event_id','sta_code','origin_DateTime']].head(10)))

		# If there are overlapping metadata windows, more than one record could be applied. This selects the first matching station.
		found_pol_df=found_pol_df.drop_duplicates(subset=['sta_code','event_id','source']).reset_index(drop=True)

		# Makes sure the number records with found and missing metadata match the original number of records
		if (len(found_pol_df)+len(missing_pol_df))!=len(pol_df):
			raise ValueError('Unexpected error while matching station start/times with polarities. The number of found+missing polarities do not match the original number of polarities.')

		# Ignores the start/end times and appends the first matching metadata record for each measurement
		missing_pol_df=missing_pol_df.merge(station_df.drop_duplicates(subset=['sta_code']),on='sta_code',how='left')

		# Concats the found/previously missing records back together
		pol_df=pd.concat([found_pol_df,missing_pol_df]).sort_values(by=['event_id','sta_code']).reset_index(drop=True)

		print('Measurements merged with station metadata considering start/end times. Runtime: {:.2f} sec'.format(time.time()-time_merge_event_sta))

	else:

		# Looks to see if there are multiple locations for the same metadata selection. If so, the first listed location is used
		duplicate_station_index=station_df[station_df.duplicated(subset='sta_code')].index
		if not(duplicate_station_index.empty):
			print(('*WARNING: {} station records have duplicate locations for your selected metadata attributes: {}\n'+
							'\tThe first station location will be used. Example problematic station info:\n{}\n').
							format(len(duplicate_station_index),p_dict['merge_on'],station_df.loc[duplicate_station_index,:].reset_index(drop=True).head(10).to_string()))
			station_df=station_df.drop(duplicate_station_index).reset_index(drop=True)

		pol_df=pol_df.merge(station_df[['sta_code','station_lat','station_lon','station_depth_km']],on='sta_code',how='left')

	missing_metadata_index=pol_df[pd.isnull(pol_df.loc[:,['station_lat','station_lon','station_depth_km']]).any(axis=1)].index
	if len(missing_metadata_index)>0:
		missing_sta_df=pol_df.loc[missing_metadata_index,'sta_code'].drop_duplicates().reset_index(drop=True)
		raise ValueError('Error: Missing metadata for {} stations:\n{}'.format(len(missing_sta_df),missing_sta_df.to_string()))

	# Calculates source-receiver distances
	aspect=np.cos(np.deg2rad(pol_df['origin_lat'].values))
	dx=(pol_df['station_lon'].values-pol_df['origin_lon'].values)*111.2*aspect
	dy=(pol_df['station_lat'].values-pol_df['origin_lat'].values)*111.2
	pol_df['sr_dist_km']=np.sqrt(dx**2+dy**2)

	return pol_df


def read_reverse_file(p_dict):
	'''
	Reads input station reversal file.
	'''
	if (p_dict['input_format']=='skhash'):
		pol_reverse_df=read_reverse_skhash_file(p_dict['plfile'],p_dict['merge_on'])
	elif (p_dict['input_format'][:-1]=='hash'):
		pol_reverse_df=read_reverse_hash_file(p_dict['plfile'])
	else:
		raise ValueError('Unknown station polarity reversal file format ({}).'.format(p_dict['input_format']))
	return pol_reverse_df


def read_reverse_skhash_file(plfile,merge_on):
	'''
	Reads input station reversal file using the SKHASH format.
	'''
	try:
		pol_reverse_df=pd.read_csv(plfile,skipinitialspace=True,comment='#',index_col=False)
	except pd.errors.EmptyDataError:
		print('The polarity reversal file (plfile: {}) is empty.'.format(plfile))
		pol_reverse_df=pd.DataFrame(columns=np.hstack([merge_on,'start_time','end_time']))

	# Replaces null locations with empty string
	if 'location' in pol_reverse_df.columns:
		pol_reverse_df['location']=pol_reverse_df['location'].fillna('')

	required_column_names=['start_time','end_time']
	if not(set(required_column_names).issubset(pol_reverse_df.columns)):
		raise ValueError('When using the skhash input format for station metadata files, the following header names MUST be provided:\n{}'.format(required_column_names))
	for req_col in merge_on:
		if not(req_col in pol_reverse_df.columns):
			raise ValueError('When using SKHASH input format and require_{}_match==True, the station polarity reversal file (plfile: {}) must contain the column "require_{}_match".'.format(req_col,plfile,req_col))

	pol_reverse_df.loc[(pol_reverse_df['start_time'].astype(str)=='0'),'start_time']='1900-01-01'
	pol_reverse_df.loc[(pol_reverse_df['end_time'].astype(str)=='0'),'end_time']='2200-01-01'

	pol_reverse_df['start_time']=pd.to_datetime(pol_reverse_df['start_time'])
	pol_reverse_df['end_time']=pd.to_datetime(pol_reverse_df['end_time'])
	return pol_reverse_df


def read_reverse_hash_file(plfile):
	'''
	Reads input station reversal file using the HASH format.
	'''
	pol_reverse_df=pd.read_csv(plfile,sep='\\s+',names=['station','start_time_int','end_time_int'])
	pol_reverse_df.loc[pol_reverse_df['start_time_int']==0,'start_time_int']=19000101
	pol_reverse_df.loc[pol_reverse_df['end_time_int']==0,'end_time_int']=22000101
	pol_reverse_df['start_time']=pd.to_datetime(pol_reverse_df['start_time_int'],format='%Y%m%d')
	pol_reverse_df['end_time']=pd.to_datetime(pol_reverse_df['end_time_int'],format='%Y%m%d')
	pol_reverse_df=pol_reverse_df.drop(columns=['start_time_int','end_time_int'])
	return pol_reverse_df


def reverse_polarities(pol_df,pol_reverse_df,p_dict):
	'''
	Flips the polarities for reversed stations.
	'''
	if (not(pol_df.empty)) & (not(pol_reverse_df.empty)):
		if (p_dict['input_format']=='skhash'):

			if not('origin_DateTime' in pol_df.columns):
				print('Applying station reversals ignoring the origin time.')

			# Computes "sta codes" by combining the desired metadata (network, station, location, channel) codes
			pol_reverse_df['sta_code']=pol_reverse_df[p_dict['merge_on']].agg('.'.join, axis=1)

			# Only considers reversed stations that have a polarity measurement
			pol_reverse_df=pol_reverse_df.loc[pol_reverse_df['sta_code'].isin(pol_df['sta_code']),:].reset_index(drop=True)

			# Determines the indecies to be reversed
			potential_reverse_ind=pol_df.index[pol_df['sta_code'].isin(pol_reverse_df['sta_code'])]
			potential_reverse_sta_code=pol_df.loc[potential_reverse_ind,'sta_code'].unique()
			group_potential_reverse_pol_df=pol_df.loc[potential_reverse_ind,:].groupby('sta_code')
			reverse_indicies=[]
			for sta_code in potential_reverse_sta_code:
				tmp_df=group_potential_reverse_pol_df.get_group(sta_code)
				sta_pol_reverse_df=pol_reverse_df.loc[pol_reverse_df['sta_code']==sta_code,:].reset_index(drop=True)

				if 'origin_DateTime' in tmp_df.columns:
					for tmp_x in range(len(sta_pol_reverse_df)):
						reverse_indicies.append(tmp_df.index[ ((tmp_df['origin_DateTime']>sta_pol_reverse_df.loc[tmp_x,'start_time']) & (tmp_df['origin_DateTime']<sta_pol_reverse_df.loc[tmp_x,'end_time'])) ])
				else:
					reverse_indicies.append(tmp_df.index)
			if len(reverse_indicies)>0:
				reverse_indicies=np.unique(np.hstack(reverse_indicies))

			pol_df.loc[reverse_indicies,'p_polarity']=pol_df.loc[reverse_indicies,'p_polarity']*-1

		elif (p_dict['input_format'][:-1]=='hash'):
			# Only considers reversed stations that have a polarity measurement
			pol_reverse_df=pol_reverse_df.loc[pol_reverse_df['station'].isin(pol_df['station']),:].reset_index(drop=True)

			# Determines the indecies to be reversed
			potential_reverse_ind=pol_df.index[pol_df['station'].isin(pol_reverse_df['station'])]
			potential_reverse_sta=pol_df.loc[potential_reverse_ind,'station'].unique()
			group_potential_reverse_pol_df=pol_df.loc[potential_reverse_ind,:].groupby('station')
			reverse_indicies=[]
			for sta in potential_reverse_sta:
				tmp_df=group_potential_reverse_pol_df.get_group(sta)
				sta_pol_reverse_df=pol_reverse_df.loc[pol_reverse_df['station']==sta,:].reset_index(drop=True)
				for tmp_x in range(len(sta_pol_reverse_df)):
					reverse_indicies.append(tmp_df.index[ ((tmp_df['origin_DateTime']>sta_pol_reverse_df.loc[tmp_x,'start_time']) & (tmp_df['origin_DateTime']<sta_pol_reverse_df.loc[tmp_x,'end_time'])) ])
			reverse_indicies=np.unique(np.hstack(reverse_indicies))

			pol_df.loc[reverse_indicies,'p_polarity']=pol_df.loc[reverse_indicies,'p_polarity']*-1

		else:
			raise ValueError('Unknown station polarity reversal file format ({}).'.format(p_dict['input_format']))

	return pol_df

def dm2dd_vec(deg, min, hemisphere):
	'''
	Converts degrees-minutes to decimal degrees
	'''
	deg=np.asarray(deg).astype(float)
	min=np.asarray(min).astype(float)
	dd = deg + min/60
	tmp_ind=np.where((hemisphere == 'W') | (hemisphere == 'S'))[0]
	dd[tmp_ind] *= -1
	return dd


'''
Functions for reading and computing S/P amplitude ratios.
'''

# External libraries
import numpy as np
import pandas as pd


def read_amp_corr_files(ampfile_in,p_dict):
    '''
    Reads the S/P ratio file (ampfile) and applies the corrections (corfile)
    '''
    # Reads S/P file and calculates ratios
    spamp_df=read_amp_file(ampfile_in,p_dict['input_format'],p_dict['merge_on'],p_dict['ratmin'],p_dict['min_sp'],p_dict['max_sp'])

    if spamp_df.empty:
        p_dict['ampfile']=''
    else:
        # Reads station corrections and applies them to the S/P measurements
        if p_dict['corfile']:
            stacor_df=read_sta_corr(p_dict)
            spamp_df=apply_sta_correction(spamp_df,stacor_df)

    return spamp_df,p_dict


def read_sta_corr(p_dict):
    '''
    Reads file of station corrections. Returns a data frame containing the
    "sta codes" and their corresponding correction.
    '''
    if p_dict['input_format']=='skhash':
        stacor_df=read_skhash_sta_corr(p_dict['corfile'],p_dict['merge_on'])
    elif p_dict['input_format']=='hash3':
        stacor_df=read_hash3_sta_corr(p_dict['corfile'])
    else:
        raise ValueError('Unknown S/P amplitude station correction file file format ({}).'.format(p_dict['input_format']))

    # Creates "sta codes" for the corrections
    stacor_df['sta_code']=stacor_df[p_dict['merge_on']].agg('.'.join, axis=1)

    # Checks for any duplicate station correction entries
    stacor_dup_flag=stacor_df['sta_code'].duplicated()
    if stacor_dup_flag.any():
        print('*WARNING: There are {} duplicate station corrections. Only the first occurrence of each will be kept. Example duplicate:\n{}'.
            format(np.sum(stacor_dup_flag),stacor_df.loc[[np.argmax(stacor_dup_flag)],:].to_string()))
        stacor_df=stacor_df.drop(stacor_df[stacor_dup_flag].index).reset_index(drop=True)

    stacor_df=stacor_df[['sta_code','sta_correction']]

    return stacor_df


def read_hash3_sta_corr(corfile):
    '''
    Reads file of station corrections using the HASH3 format
    '''
    stacor_df=pd.read_csv(corfile,sep='\\s+',names=['station','channel','network','sta_correction'])
    return stacor_df


def read_skhash_sta_corr(corfile,merge_on):
    '''
    Reads file of station corrections using the SKHASH format
    '''
    consider_cols=['network','station','location','channel','sta_correction']
    try:
        stacor_df=pd.read_csv(corfile,skipinitialspace=True,comment='#',usecols=lambda x: x in consider_cols)
    except pd.errors.EmptyDataError:
        print('*Warning: corfile ({}) is empty.'.format(corfile))
        stacor_df=[]
    if not('sta_correction' in stacor_df.columns):
        raise ValueError('When using SKHASH input format, the corfile ({}) must contain the column "sta_correction".'.format(corfile))
    for req_col in merge_on:
        if not(req_col in stacor_df.columns):
            raise ValueError('When using SKHASH input format and require_{}_match==True, the corfile ({}) must contain the column "{}".'.format(req_col,corfile,req_col))
    return stacor_df


def read_amp_file(ampfile,input_format,merge_on,ratmin,min_sp,max_sp):
    '''
    Reads file of S/P ratios. Returns a data frame containing the
    "sta codes" and their corresponding S/P ratios.
    '''
    if input_format=='skhash':
        spamp_df=read_skhash_amp_file(ampfile,merge_on)
    elif input_format=='hash3':
        spamp_df=read_hash3_amp_file(ampfile)
    else:
        raise ValueError('Unknown S/P amplitude file file format ({}).'.format(input_format))

    spamp_df['event_id']=spamp_df['event_id'].astype(str)
    if 'event_id2' in spamp_df.columns:
        spamp_df['event_id2']=spamp_df['event_id2'].astype(str)

    if ('sp_ratio' in spamp_df.columns):
        spamp_df['sp_ratio']=spamp_df['sp_ratio'].abs()
    else:
        # spamp_df.loc[:,['noise_p','noise_s','amp_p','amp_s']]=spamp_df.loc[:,['noise_p','noise_s','amp_p','amp_s']].abs()
        spamp_df[['noise_p','noise_s','amp_p','amp_s']]=spamp_df[['noise_p','noise_s','amp_p','amp_s']].abs()

        spamp_df=spamp_df.loc[((spamp_df['amp_p']/spamp_df['noise_p'])>=ratmin) &
                                ((spamp_df['amp_s']/spamp_df['noise_s'])>=ratmin),:].reset_index(drop=True)
        spamp_df['sp_ratio']=(spamp_df['amp_s']/spamp_df['amp_p']).round(2)
        spamp_df=spamp_df.drop(columns=['noise_p','noise_s','amp_p','amp_s'])

    spamp_df=spamp_df.drop(spamp_df[spamp_df['sp_ratio']==0].index).reset_index(drop=True)

    # Sets the minimum and maximum S/P ratios
    if min_sp>0:
        spamp_df.loc[spamp_df.sp_ratio<min_sp,'sp_ratio']=min_sp
    if max_sp>0:
        spamp_df.loc[spamp_df.sp_ratio>max_sp,'sp_ratio']=max_sp

    spamp_df['sp_ratio']=np.log10(spamp_df['sp_ratio'])
    spamp_df['sta_code']=spamp_df[merge_on].agg('.'.join, axis=1)

    spamp_df=spamp_df.filter(['event_id','event_id2','sta_code','sp_ratio','origin_lat','origin_lon','origin_depth_km','takeoff','takeoff_uncertainty','azimuth','azimuth_uncertainty'])

    return spamp_df


def read_skhash_amp_file(ampfile,merge_on):
    '''
    Reads file of S/P ratios using the SKHASH format
    '''
    consider_cols=['event_id','event_id2','network','station','location','channel','noise_p','noise_s','amp_p','amp_s','sp_ratio','origin_latitude','origin_longitude','origin_depth_km','takeoff','takeoff_uncertainty','azimuth','azimuth_uncertainty']
    try:
        spamp_df=pd.read_csv(ampfile,skipinitialspace=True,usecols=lambda x: x in consider_cols)
    except pd.errors.EmptyDataError:
        print('*Warning: ampfile ({}) is empty.'.format(ampfile))
        spamp_df=[]
    if not('event_id' in spamp_df.columns):
        raise ValueError('When using SKHASH input format, the ampfile ({}) must contain the column "event_id".'.format(ampfile))
    if (not({'noise_p','noise_s','amp_p','amp_s'}.issubset(spamp_df.columns))) and (not('sp_ratio' in spamp_df.columns)):
        raise ValueError('When using SKHASH input format, the ampfile ({}) must either contain one of the following sets of column names:\n\t1. noise_p,noise_s,amp_p,amp_s\n\t2. sp_ratio'.format(ampfile))

    for req_col in merge_on:
        if not(req_col in spamp_df.columns):
            raise ValueError('When using SKHASH input format and require_{}_match==True, the ampfile ({}) must contain the column "{}".'.format(req_col,ampfile,req_col))

    spamp_df=spamp_df.rename(columns={'origin_latitude':'origin_lat','origin_longitude':'origin_lon'})

    return spamp_df


def read_hash3_amp_file(ampfile):
    '''
    Reads file of S/P amplitude ratios using the HASH3 format
    '''
    file1 = open(ampfile, 'r')
    event_amplitude_list=[]
    sta=[];net=[];loc=[];cha=[];noise_p=[];noise_s=[];amp_p=[];amp_s=[]
    event_id=''
    for line in file1:
        split_line=line.split()
        if len(split_line)==2: # Event id line
            if len(sta)>0:
                tmp_df=pd.DataFrame(data=np.asarray(np.vstack((sta,net,loc,cha,noise_p,noise_s,amp_p,amp_s))).T,
                                    columns=['station','network','location','channel','noise_p','noise_s','amp_p','amp_s'])
                tmp_df['event_id']=event_id
                tmp_df=tmp_df.astype({'noise_p':float,'noise_s':float,'amp_p':float,'amp_s':float})
                event_amplitude_list.append(tmp_df)
                sta=[];net=[];loc=[];cha=[];noise_p=[];noise_s=[];amp_p=[];amp_s=[]
            event_id=split_line[0]
        else: # Amplitude line
            sta.append(split_line[0])
            cha.append(split_line[1])
            net.append(split_line[2])
            noise_p.append(split_line[5])
            noise_s.append(split_line[6])
            amp_p.append(split_line[7])
            amp_s.append(split_line[8])
            loc.append('--')
    if len(sta)>0:
        tmp_df=pd.DataFrame(data=np.asarray(np.vstack((sta,net,loc,cha,noise_p,noise_s,amp_p,amp_s))).T,
                            columns=['station','network','location','channel','noise_p','noise_s','amp_p','amp_s'])
        tmp_df['event_id']=event_id
        tmp_df=tmp_df.astype({'noise_p':float,'noise_s':float,'amp_p':float,'amp_s':float})
        event_amplitude_list.append(tmp_df)
    file1.close()

    if len(event_amplitude_list)>0:
        spamp_df=pd.concat(event_amplitude_list).reset_index(drop=True)
    else:
        spamp_df=pd.DataFrame(columns=['station','network','location','channel','noise_p','noise_s','amp_p','amp_s'])

    return spamp_df


def apply_sta_correction(spamp_df,stacor_df):
    '''
    Applies station corrections to S/P ratios.
    '''
    spamp_df=spamp_df.merge(stacor_df,on='sta_code',how='left')
    spamp_df.loc[spamp_df[pd.isnull(spamp_df['sta_correction'])].index,'sta_correction']=0
    spamp_df['sp_ratio']-=spamp_df['sta_correction']
    spamp_df=spamp_df.drop(columns='sta_correction')

    return spamp_df



# External libraries
import numpy as np
import pandas as pd


def create_outfile1(outfile1,cat_df,pol_df):
    '''
    Creates the outfile of the preferred mech solutions
    '''
    header_str_out='event_id,strike,dip,rake,quality,fault_plane_uncertainty,aux_plane_uncertainty,num_p_pol,num_sp_ratios,'+\
                        'azimuthal_gap,takeoff_gap,polarity_misfit,prob_mech,sta_distribution_ratio,sp_misfit,mult_solution_flag'
    if len(cat_df):
        header_str_out+=','
        if 'origin_DateTime' in cat_df:
            header_str_out+='time,'
        if 'event_mag' in pol_df:
            header_str_out+='magnitude,'
        header_str_out+='origin_lat,origin_lon,origin_depth_km,horz_uncert_km,vert_uncert_km'
    with open(outfile1, "w") as f_outfile1:
        f_outfile1.write(header_str_out+'\n')
    return True


def create_outfile2(outfile2):
    '''
    Creates the outfile of the acceptable mech solutions
    '''
    with open(outfile2, "w") as f_outfile2:
        f_outfile2.write('event_id,mech_number,strike,dip,rake,norm_N,norm_E,norm_Z,norm_aux_N,norm_aux_E,norm_aux_Z\n')
    return True


def write_outfile1(outfile1,mech_df,event_df,event_id):
    '''
    Writes mech solutions for an event to outfile1.
    '''
    event_str_output=''
    if len(event_df):
        event_str_output+=','
        if 'origin_DateTime' in event_df:
            event_str_output+=event_df.origin_DateTime.strftime('%Y-%m-%d %X')+','
        if 'event_mag' in event_df:
            event_str_output+=str(event_df['event_mag'])+','
        event_str_output+=str(event_df['origin_lat'])+','
        event_str_output+=str(event_df['origin_lon'])+','
        event_str_output+=str(event_df['origin_depth_km'])+','
        event_str_output+=str(event_df['horz_uncert_km'])+','
        event_str_output+=str(event_df['vert_uncert_km'])

    with open(outfile1, "a") as f_outfile1:
        if len(mech_df)>1:
            mech_df['mflag']=True
        else:
            mech_df['mflag']=False
        for imult in range(len(mech_df)):
            f_outfile1.write(('{},{},{},{},{},'+
                              '{},{},{},{},{},'+
                              '{},{},{},{},{},{}{}\n').format(
                event_id,
                mech_df.loc[imult,'str_avg'], # strike
                mech_df.loc[imult,'dip_avg'], # dip
                mech_df.loc[imult,'rak_avg'], # rake
                mech_df.loc[imult,'qual'], # ad-hoc mech quality
                mech_df.loc[imult,'rms_diff'], # fault plane uncertainty
                mech_df.loc[imult,'rms_diff_aux'], # aux plane uncertainty
                mech_df.loc[imult,'num_p_pol'], # num p polarity picks
                mech_df.loc[imult,'num_sp_ratios'], # num S/P ratios
                mech_df.loc[imult,'azimuthal_gap'], # max azimuthal gap between measurements
                mech_df.loc[imult,'takeoff_gap'], # max takeoff gap between measurements
                mech_df.loc[imult,'mfrac'], # weighted percent misfit of first motions
                mech_df.loc[imult,'prob'], # probability mechanism close to solution
                mech_df.loc[imult,'stdr'], # 100*(station distribtuion ratio)
                mech_df.loc[imult,'mavg'], # 100*(average log10(S/P) misfit)
                mech_df.loc[imult,'mflag'], # Flag indicating whether there are multiple solutions for the event
                event_str_output
                ))
    return True


def write_outfile2(outfile2,event_id,strike_all,dip_all,rake_all,faultnorms_all,faultslips_all,output_angle_precision,output_vector_precision):
    '''
    Writes acceptable mech solutions for an event to outfile2.
    '''
    acceptable_mechs_df=pd.DataFrame(columns=['event_id','mech_number','strike','dip','rake',
                                              'norm_N','norm_E','norm_Z','norm_aux_N','norm_aux_E','norm_aux_Z'],
                                              index=np.arange(len(strike_all)))

    acceptable_mechs_df['event_id']=event_id
    acceptable_mechs_df['mech_number']=np.arange(len(acceptable_mechs_df))
    acceptable_mechs_df['strike']=np.round(strike_all,output_angle_precision)
    acceptable_mechs_df['dip']=np.round(dip_all,output_angle_precision)
    acceptable_mechs_df['rake']=np.round(rake_all,output_angle_precision)
    acceptable_mechs_df['norm_N']=np.round(faultnorms_all[0,:],output_vector_precision)
    acceptable_mechs_df['norm_E']=np.round(faultnorms_all[1,:],output_vector_precision)
    acceptable_mechs_df['norm_Z']=np.round(faultnorms_all[2,:],output_vector_precision)
    acceptable_mechs_df['norm_aux_N']=np.round(faultslips_all[0,:],output_vector_precision)
    acceptable_mechs_df['norm_aux_E']=np.round(faultslips_all[1,:],output_vector_precision)
    acceptable_mechs_df['norm_aux_Z']=np.round(faultslips_all[2,:],output_vector_precision)
    acceptable_mechs_df.to_csv(outfile2,mode='a',sep=',',index=False,header=False)
    return True


def pol_agree(pol_df,p_dict):
    '''
    Writes the polarity agreement records for all events to outfile_pol_agree
    '''
    # Selects only measurements with a mech
    pol_df=pol_df.loc[~pd.isnull(pol_df['mech_quality'])].reset_index(drop=True)

    if (p_dict['min_quality_report']):
        pol_df=pol_df[pol_df.mech_quality<=p_dict['min_quality_report']].reset_index(drop=True)

    if not(pol_df.empty):
        pol_df['pol_agreement_weighted']=pol_df['pol_agreement']*(pol_df['p_polarity'].abs())

        group_pol_df=pol_df.groupby(['sta_code','source'])

        pol_agree_df=group_pol_df.agg({'pol_agreement':['sum','count'],'p_polarity':[lambda lam: (abs(lam).sum())],'pol_agreement_weighted':['sum'] })
        pol_agree_df['count_correct']=pol_agree_df['pol_agreement']['sum'].astype(int)
        pol_agree_df['count_total']=pol_agree_df['pol_agreement']['count'].astype(int)
        pol_agree_df['pol_accuracy']=(100*pol_agree_df['count_correct']/pol_agree_df['count_total']).round(1)

        pol_agree_df['weight_correct']=pol_agree_df['pol_agreement_weighted']['sum'].astype(float)
        pol_agree_df['weight_total']=pol_agree_df['p_polarity']['<lambda>'].astype(float)
        pol_agree_df['weight_pol_accuracy']=(100*pol_agree_df['weight_correct']/pol_agree_df['weight_total']).round(1)
        pol_agree_df['weight_correct']=pol_agree_df['weight_correct'].round(5)
        pol_agree_df['weight_total']=pol_agree_df['weight_total'].round(5)

        pol_agree_df=pol_agree_df.drop(columns=['pol_agreement','p_polarity','pol_agreement_weighted'])
        pol_agree_df=pol_agree_df.reset_index()
        pol_agree_df.columns=pol_agree_df.columns.droplevel(1)
        pol_agree_df=pol_agree_df.sort_values(by=['weight_pol_accuracy','count_correct','sta_code'],ascending=[True,False,True])

        pol_agree_df=pol_agree_df.loc[pol_agree_df.count_total>0,:].reset_index(drop=True)

        pol_agree_df.to_csv(p_dict['outfile_pol_agree'],index=False)
    return True


def sp_agree(pol_df,p_dict):
    '''
    Writes the S/P agreement records for all events to outfile_sp_agree
    '''
    if (p_dict['min_quality_report']):
        pol_df=pol_df[pol_df.mech_quality<=p_dict['min_quality_report']]
    if not(pol_df.empty):

        group_sp_agree_df=pol_df.groupby(['sta_code','source'])

        sp_agree_df=group_sp_agree_df['sp_diff'].agg(['mean','std','count']).reset_index()
        sp_agree_df=sp_agree_df.drop(np.where(sp_agree_df['count']==0)[0])

        sp_agree_df=sp_agree_df.sort_values(by=['mean','count','std'],ascending=[True,False,True])
        sp_agree_df.loc[pd.isnull(sp_agree_df['std']),'std']=-999.

        sp_agree_df['mean']=(sp_agree_df['mean']).round(4)
        sp_agree_df['std']=(sp_agree_df['std']).round(4)
        sp_agree_df=sp_agree_df.rename(columns={'mean':'avg_sp_diff','std':'std_sp_diff'})

        sp_agree_df.to_csv(p_dict['outfile_sp_agree'],index=False)
    return True


def pol_info(pol_df,p_dict):
    '''
    Writes information about the pol info for all events to outfile_pol_info
    '''
    if 'azimuth' in pol_df.columns:
        pol_df['azimuth']=pol_df['azimuth'].round(p_dict['output_angle_precision'])
    if 'takeoff' in pol_df.columns:
        pol_df['takeoff']=pol_df['takeoff'].round(p_dict['output_angle_precision'])
    if 'sp_ratio' in pol_df.columns:
        pol_df['sp_ratio']=(10**pol_df['sp_ratio']).round(5)
    if 'sp_diff' in pol_df.columns:
        pol_df['sp_diff']=(10**pol_df['sp_diff']).round(5)
    if 'event_mag' in pol_df.columns:
        pol_df['event_mag']=pol_df['event_mag'].round(5)

    if not(pol_df.empty):
        pol_df.to_csv(p_dict['outfile_pol_info'],index=False)
    return True



'''
Python package for earthquake focal mechanism inversions.

Author: Robert J. Skoumal (rskoumal@usgs.gov)

This project contains Python code to compute focal mechanism solutions using
first-motion polarities (traditional, consensus, and/or imputed) and S/P ratios
(traditional and/or consensus).

Significant portions of this code are based on the HASH algorithm created by
Jeanne L. Hardebeck and Peter M. Shearer. If you use this code, please cite the
appropriate papers.
	Skoumal, R.J., Hardebeck, J.L., & Shearer, P.M. (2024). SKHASH: A Python
		package for computing earthquake focal mechanisms. Seismological Research Letters,
		95(4), 2519-2526. https://doi.org/10.1785/0220230329
	Hardebeck, J.L., & Shearer, P.M. (2002). A new method for determining
		first-motion focal mechanisms. Bulletin of the Seismological Society of America,
		92(6), 2264-2276. https://doi.org/10.1785/0120010200
	Hardebeck, J.L., & Shearer, P.M. (2003). Using S/P amplitude ratios to
		constrain the focal mechanisms of small earthquakes. Bulletin of the Seismological
		Society of America, 93(6), 2434-2444. https://doi.org/10.1785/0120020236
'''

# Standard libraries
import sys
import time
import importlib.util
import multiprocessing
import argparse

# External libraries
import numpy as np
import pandas as pd

# Local libraries
import functions.in_pol as in_pol # Reading polarity inputs
import functions.in_sta as in_sta # Reading station files
import functions.in_sp as in_sp # Reading S/P ratios
import functions.in_qc as in_qc # Quality control of user inputs
import functions.in_other as in_other # Functions for reading other inputs
import functions.fun as fun # Computing mechanisms
import functions.out as out # Output functions
import functions.gridsearch_so as gridsearch_so # For creating fortran gridsearch module
import functions.compute_mech as compute_mech # For computing mechanism

def SKHASH():
	# Superficial version information
	version_string='v1.1'
	version_date='2025-05-13'

	'''
	Sets default parameter values. If a value is not provided in the control file (or command line) for a variable, the default value is used.
	'''
	p_dict={
		'input_format':'skhash', # default input file format
		'input_format_stfile':'', # station list file format
		'input_format_fpfile':'', # traditional P-polarity file format
		'input_format_impfile':'', # imputed P-polarity input
		'input_format_conpfile':'', # consensus P-polarities file format
		'input_format_dlpfile':'', # deep learning P-polarities file format
		'input_format_ampfile':'', # amplitude input file format
		'input_format_relampfile':'', # relative S/P ratio file format
		'input_format_plfile':'', # polarity reversal format

		'controlfile':'',
		'catfile':'', # earthquake catalog information filename
		'stfile':'', # station list filename
		'plfile':'', # station polarity reversal filename
		'corfile':'', # station correction filename
		'fpfile':'', # traditional P-polarity input filename
		'impfile':'', # imputed P-polarity input filename
		'conpfile':'', # consensus P-polarities from relative measurements input filename
		'dlpfile':'', # deep learning P-polarities input filename
		'ampfile':'', # amplitude input filename
		'relampfile':'', # relative S/P ratios input filename
		'simulpsfile':'', # filename for SIMULPS azimuth and takeoff angles
		'outfile1':'default_out1.txt', # focal mechanisms output filename
		'outfile2':'', # acceptable plane output filename
		'outfile_pol_agree':'', # record of polarity (dis)agreeement output filename
		'outfile_sp_agree':'', # record of S/P difference output filename
		'outfile_pol_info':'', # record of all polarities considered in the mechanisms
		'outfolder_plots':'', # Folder where simple focal mechanism plots will be created (outfolder_plots/event_id.png). To ignore, leave blank.

		'npolmin':8, # mininum number of polarity data (e.g., 8)
		'nmc':30, # number of trials (e.g., 30)
		'maxout':500, # the maximum number of acceptable focal mechansims considered (e.g., 500).
		'ratmin':3.0, # minimum allowed signal-to-noise ratio
		'badfrac':0.1, # fraction polarities assumed bad
		'badmin':2.0, # minimum number of polarities per event assumed bad. If badmin > [(sum of the polarity weights) * badfrac], badmin replaces badfrac
		'qbadfrac':0.3, # assumed noise in amplitude ratios, log10 (e.g. 0.3 for a factor of 2)
		'qbadmin':2.0, # minimum S/P amplitude misfit allowed per event. If qbadmin > [(sum of the number of S/P measurements) * qbadfrac], qbadmin replaces qbadfrac
		'min_polarity_weight':0.1, # Any polarities with an abs(weight) < min_polarity_weight will be ignored
		'delmin':0.0, # minimum allowed source-station distance in km. If delmin<=0, uses all results >= than look_del[0] distance.
		'delmax':120.0, # maximum allowed source-station distance in km. If delmax<=0, uses all results <= than look_del[1] distance.
		'azmax':10.0, # maximum allowed source-station azimuth uncertainty in degrees. To use all results regardless of azimuthal uncertainty, set azmax=0.
		'pmax':0.0, # maximum allowed takeoff ("plungal") uncertainty in degrees. To use all results regardless of takeoff uncertainty, set pmax=0.
		'cangle':45.0, # angle for computing mechanisms probability
		'prob_max':0.2, # probability threshold for multiples (e.g., 0.2)
		'max_agap':90.0, # maximum azimuthal gap
		'max_pgap':60.0, # maximum "plungal" gap

		'vmodel_paths':[], # list of paths to velocity model
		'write_lookup_table':False, # The lookup table will be saved so that it can be used in future runs. A .npy file extension will be added to the vmodel_paths.
		'recompute_lookup_table':True, # If a velocity model with an identical filename already exists, it will remake it.

		'flip_takeoff':False, # If True, takeoff angles read/written will correspond to [180: upgoing and 0: downgoing]. If False, it will follow HASH notation [0: upgoing, 180: downgoing].

		'require_temporal_match':False, # Requires the reported earthquake to occur between the station metadata start-end time.
		'require_network_match':True, # Requires the reported pick network to match the metadata
		'require_station_match':True, # Requires the reported pick station to match the metadata
		'require_channel_match':True, # Requires the reported pick channel to match the metadata
		'require_location_match':True, # Requires the reported pick location to match the metadata
		'allow_duplicate_stations':False, # Allows an event_id to have multiple polarities and S/P ratios from the same receiver
		'remove_duplicate_stations':False, # If an event_id has multiple polarities or S/P ratios from the same receiver, removes the duplicates
		'iterative_avg':False, # If True, it will iteratively compute the average mech by removing the solution furthest from the avg following HASH.

		'min_quality_report':'', # Only mech qualities of this or better will be used to reported. To consider all accepted solutions, leave blank.

		'overwrite_output_file':True, # If True, SKHASH will overwrite any existing output files.
		'perturb_epicentral_location':False, # If True, it will randomly perturb the horizontal location of the earthquakes

		'ignore_missing_metadata':False, # If True, any measurements with missing metadata will be ignored. If False, an error will be raised.

		'use_fortran':False, # If True, it will attempt to load the Fortran subroutine. If it cannot load or is False, the Python routine will be used.
		'num_cpus':1, # Number of cores to run things in parallel. Set to 0 to use all available, or 1 to run in serial.

		'dang':5, # minimum grid spacing (degrees).
		'nx0':101, # maximum source-station distance bins for look-up tables
		'nd0':14, # maximum source depth bins for look-up tables
		'look_dep':[0,39,3], # minimum source depth, maximum, and interval for the lookup table
		'look_del':[0,200,2], # minimum source-station distance, maximum, and interval for the lookup table
		'allow_hypocenters_outside_table':False, # If a hypocenter is outside of the lookup table range, the min/max lookup value is used. If False, an error will be produced.
		'nump':9000, # number of rays traced

		'min_sp':0.0005,# If any user-provided S/P ratios are less than min_sp after applying station corrections, those values are replaced with min_sp. To ignore, use a value <= 0.
		'max_sp':2000.0, # If any user-provided S/P ratios are greater than max_sp after applying station corrections, those values are replaced with max_sp. To ignore, use a value <= 0.
		'min_amp':0.0005, # minimum amplitude [0-1] considered when calculating expected P and S amplitudes. Values < than this are treated as 0. To ignore, use a value <= 0.

		'epicenter_degree_precision':6, # Number of decimal places to round lat, lon epicenter coordinates.
		'output_angle_precision':1, # Number of decimal places to output for angles (e.g., strike, dip, rake, takeoff)
		'output_quality_precision':1, # Number of decimal places to output for mech qualities (e.g., prob, mfrac, mavg, stdr)
		'output_km_distance_precision':3, # Number of decimal places to output for source-receiver distances (reported in km)
		'output_vector_precision':4, # Number of decimal places to output for normal vectors
		'default_vert_uncert_km':1.0, # If the vertical uncertainty for an earthquake is missing or negative, this value is used instead when perturbing the source depth.
		'default_horz_uncert_km':1.0, # If the horizontal uncertainty for an earthquake is missing or negative, this value is used instead when perturbing the source hypocenter.
		'default_takeoff_uncert':0.0, # If the precomputed takeoff uncertainty is missing or negative, this value is used instead when perturbing takeoffs
		'default_azimuth_uncert':0.0, # If the precomputed azimuth uncertainty is missing or negative, this value is used instead when perturbing azimuths

		'plot_station_names':False, # If True, station names will be included in the beachball plot
		'plot_acceptable_solutions':False, # If True, all acceptable solutions will be plotted in the beachball plot
		'plot_mult_solutions':False, # If True and multiple solutions are found, all solutions will be plotted in the beachball plot

		# Warning: only change the following values if you know what you're doing! :)
		'compute_takeoff_azimuth':True, # Used to determine if takeoff and source-receiver azimuths need to be computed
		'npick0':15000, # maximum number of picks per event. Used by Fortran gridsearch. If this is changed, the value in gridsearch.f must be changed and recompiled.
		'merge_on':[] # Controls how to merge metadata with observations. List that includes ('network', 'station', 'location', 'channel')
	}

	# Quality codes should be listed from best quality (most restrictive) to worst quality (least restrictive)
	qual_criteria_dict={
					'qual_letter':np.asarray([ 'A',  'B',  'C',     'D']),
					'probs':np.asarray(      [ 0.8,  0.6,  0.5,      0 ]),
					'var_avg':np.asarray(    [  25,   35,   45, np.inf ]),
					'mfrac':np.asarray(      [0.15, 0.20, 0.30, np.inf ]),
					'stdr':np.asarray(       [ 0.5,  0.4,  0.3,      0 ]),
					# If the following conditions are not met, the lowest quality letter is given.
					'azimuthal_gap':90, # max azimuthal gap
					'takeoff_gap':60, # max takeoff gap
					'num_p_pol':8, # minimum number of p polarities
					}

	'''
	Reads command line arguments.
	Note that if you also use a control file, the control file variable will overwrite command line arguments.
	'''
	parser = argparse.ArgumentParser()
	parser.add_argument('controlfile',nargs='?')
	for p_dict_var in p_dict.keys():
		parser.add_argument('--'+p_dict_var)
	args = vars(parser.parse_args())
	for p_dict_var in p_dict.keys():
		if args[p_dict_var] is not None:
			dtype=type(p_dict[p_dict_var])
			if dtype==bool:
				if args[p_dict_var].lower() in ['true','1']:
					p_dict[p_dict_var]=True
				elif args[p_dict_var].lower() in ['false','0']:
					p_dict[p_dict_var]=False
				else:
					raise ValueError(('Expected a boolean for the command-line declared variable \'{}\''+\
						'The provided value ({}) is not a boolean.').format(p_dict_var,args[p_dict_var]))
			else:
				p_dict[p_dict_var]=dtype(args[p_dict_var])

	print('========================\nSKHASH {} ({})\n========================'.format(version_string,version_date))
	total_runtime_start=time.time()

	'''
	Checks Python version.
	'''
	if not((sys.version_info.major==3) & (sys.version_info.minor>=8)):
		print('***WARNING: SKHASH has only been tested on Python 3.8+. Refer to the manual for supported versions.')

	'''
	Reads the user-created control file, overwritting the default and commonand line parameter values.
	'''
	if p_dict['controlfile']:
		print('Control file: {}'.format(p_dict['controlfile']))
		p_dict=in_other.read_control_file(p_dict['controlfile'],p_dict)

	'''
	Loads/creates the fortran extension module
	'''
	if p_dict['use_fortran']:
		gridsearch_spec = importlib.util.find_spec("functions.gridsearch")
		if (gridsearch_spec is not None):
			import functions.gridsearch as gridsearch
			print('Fortran subroutine module successfully loaded.')
		else:
			status=gridsearch_so.create()
			if status:
				import functions.gridsearch as gridsearch
			else:
				p_dict['use_fortran']=False
				print('Unable to load Fortran subroutine. Falling back to using the Python gridsearch routine.\n')

	'''
	Performs various checks to ensure that the user-defined parameters are appropriate
	'''
	p_dict=in_qc.check_input_params(p_dict,qual_criteria_dict)

	'''
	Reads P-wave first motion polarity file(s)
	'''
	pol_df=[]
	cat_df=[]
	if p_dict['fpfile']:
		tmp_cat_df,tmp_pol_df=in_pol.read_polarity_file(p_dict['fpfile'],p_dict['input_format_fpfile'],p_dict['merge_on'])
		tmp_pol_df['source']='trad_p'
		if len(tmp_pol_df):
			pol_df.append(tmp_pol_df)
		if len(tmp_cat_df):
			cat_df.append(tmp_cat_df)
	if p_dict['impfile']:
		tmp_cat_df,tmp_pol_df=in_pol.read_polarity_file(p_dict['impfile'],p_dict['input_format_impfile'],p_dict['merge_on'])
		tmp_pol_df['source']='imp_p'
		if len(tmp_pol_df):
			pol_df.append(tmp_pol_df)
		if len(tmp_cat_df):
			cat_df.append(tmp_cat_df)
	if p_dict['conpfile']:
		tmp_cat_df,tmp_pol_df=in_pol.read_polarity_file(p_dict['conpfile'],p_dict['input_format_conpfile'],p_dict['merge_on'])
		tmp_pol_df['source']='con_p'
		if len(tmp_pol_df):
			pol_df.append(tmp_pol_df)
		if len(tmp_cat_df):
			cat_df.append(tmp_cat_df)
	if p_dict['dlpfile']:
		tmp_cat_df,tmp_pol_df=in_pol.read_polarity_file(p_dict['dlpfile'],p_dict['input_format_dlpfile'],p_dict['merge_on'])
		tmp_pol_df['source']='dl_p'
		if len(tmp_pol_df):
			pol_df.append(tmp_pol_df)
		if len(tmp_cat_df):
			cat_df.append(tmp_cat_df)
	if len(pol_df):
		pol_df=pd.concat(pol_df).reset_index(drop=True)
	else:
		pol_df=pd.DataFrame()
		print('*WARNING: No polarity measurements provided.')
	if len(cat_df):
		cat_df=pd.concat(cat_df).drop_duplicates(subset='event_id').reset_index(drop=True)

	# Adds SIMULPS source-receiver distances, takeoff, and azimuths
	if p_dict['simulpsfile']:
		pol_df=in_other.read_simulps(p_dict['simulpsfile'],pol_df)

	'''
	Reads S/P amplitude file, calculates S/P ratios, discards S/P ratios below the noise threshold, and applies station corrections.
	'''
	if p_dict['ampfile']:
		spamp_df,p_dict['ampfile']=in_sp.read_amp_corr_files(p_dict['ampfile'],p_dict)
		spamp_df['source']='trad_sp'

		# Concats polarity and S/P data into a single dataframe
		pol_df=pd.concat([pol_df,spamp_df]).reset_index(drop=True)

	if p_dict['relampfile']:
		spamp_df,p_dict=in_sp.read_amp_corr_files(p_dict['relampfile'],p_dict)
		spamp_df['source']='rel_sp'

		# Concats polarity and consensus S/P data into a single dataframe
		pol_df=pd.concat([pol_df,spamp_df]).reset_index(drop=True)

	# Looks for duplicate polarity and S/P measurements
	if not(p_dict['allow_duplicate_stations']):
		if len(pol_df):
			if 'sp_ratio' in pol_df.columns:
				dup_df=pd.concat([pol_df[['event_id','sta_code']],pd.isnull(pol_df[['p_polarity','sp_ratio']])],axis=1)
			else:
				dup_df=pd.concat([pol_df[['event_id','sta_code']],pd.isnull(pol_df[['p_polarity']])],axis=1)
			duplicate_polarity_index=pol_df[dup_df.duplicated()].index
			if not(duplicate_polarity_index.empty):
				if p_dict['remove_duplicate_stations']:
					print('Removing duplicate information for {} measurements. Example issues:\n{}'.
						format(len(duplicate_polarity_index),pol_df.loc[duplicate_polarity_index,['event_id','sta_code']].head(10).to_string()))
					pol_df=pol_df.drop(duplicate_polarity_index).reset_index(drop=True)
				else:
					raise ValueError('Duplicate information for {} measurements. Example issues:\n{}'.
						format(len(duplicate_polarity_index),pol_df.loc[duplicate_polarity_index,['event_id','sta_code']].head(10).to_string()))

	if len(pol_df)==0:
		print('No polarity information provided. Exiting.')
		quit()

	'''
	Reads earthquake catalog
	'''
	if p_dict['catfile']:
		cat_df=in_other.read_catalog_file(p_dict)

	# Adds event times, locations, and uncertainties to polarity information
	if len(cat_df):
		if not('event_id2' in pol_df.columns):
			pol_df=pol_df.merge(cat_df,on='event_id',how='left')
		else: # composite mechs
			pol_df=pol_df.merge(cat_df.rename(columns={'event_id':'event_id2'}),on='event_id2',how='left')

	'''
	Quality controls the datasets
	'''
	# Quality control of catalog dataset
	cat_df=in_qc.check_cat_df(cat_df,pol_df,p_dict)

	# Quality control of polarity dataset
	pol_df=in_qc.check_pol_df(pol_df,cat_df,p_dict)

	'''
	Checks that earthquake locations will be contained in the model
	'''
	if p_dict['compute_takeoff_azimuth']:
		pol_df,p_dict=in_qc.check_model_eq(cat_df,pol_df,p_dict)

	'''
	Reads station polarity reversals and applies them to the polarity measurements
	'''
	if p_dict['plfile']:
		pol_reverse_df=in_sta.read_reverse_file(p_dict)
		pol_df=in_sta.reverse_polarities(pol_df,pol_reverse_df,p_dict)
	pol_df=pol_df.drop(pol_df.filter(['station']),axis=1)

	'''
	Reads station metadata file and appends the locations to the polarities
	'''
	if p_dict['stfile']:
		station_df=in_sta.read_station_file(pol_df,p_dict)
		pol_df=in_sta.apply_station_locations(pol_df,station_df,p_dict)

	# Drops columns that are no longer needed
	if 'origin_DateTime' in pol_df:
		pol_df=pol_df.drop(columns=['origin_DateTime'])

	'''
	Dropping any measurements with source-receiver distances > delmax
	'''
	pol_df['sr_dist_km']=pol_df['sr_dist_km'].round(p_dict['output_km_distance_precision'])
	if (p_dict['delmax']>0) | (p_dict['delmin']>0):
		drop_flag=((pol_df['sr_dist_km']>p_dict['delmax']) | (pol_df['sr_dist_km']<p_dict['delmin']))
		if np.any(drop_flag):
			if p_dict['delmin']>0:
				print('Discarded {} of the {} polarity measurements with source-receiver distances <{} or >{} km'.
					format(np.sum(drop_flag),len(pol_df),p_dict['delmin'],p_dict['delmax']))
			else:
				print('Discarded {} of the {} polarity measurements with source-receiver distances >{} km'.
					format(np.sum(drop_flag),len(pol_df),p_dict['delmax']))
			pol_df=pol_df.drop(pol_df[drop_flag].index).reset_index(drop=True)

	'''
	Calculates maximum possible source-receiver azimuth variation, dropping any polarities > azmax
	'''
	if p_dict['azmax']>0:
		if not('azimuth_uncertainty' in pol_df.columns):
			sr_unc_flag=pol_df['horz_uncert_km']<=pol_df['sr_dist_km']

			pol_df['azimuth_uncertainty']=(90-np.rad2deg(np.arccos(pol_df['horz_uncert_km']/pol_df['sr_dist_km'].values)))*2

			# If the horizontal uncertainty is greater than the source-receiver distance, any azimuth is possible.
			tmp_index=pol_df[~sr_unc_flag].index
			pol_df.loc[tmp_index,'azimuth_uncertainty']=360

		# Drop polarities that have a source-receiver azimuth variation > azmax
		tmp_index=pol_df[pol_df['azimuth_uncertainty'] > p_dict['azmax']].index
		if not(tmp_index.empty):
			pol_df.drop(tmp_index, inplace=True)
			print('Discarded {} polarity measurement(s) with source-receiver azimuth uncertainties >{} deg'.format(len(tmp_index),p_dict['azmax']))

		pol_df['azimuth_uncertainty']=pol_df['azimuth_uncertainty'].round(p_dict['output_angle_precision'])
		pol_df=pol_df.reset_index(drop=True)

	'''
	If a station file is given but takeoff and azimuths are already provided, ignore the station file.
	'''
	if p_dict['stfile']:
		if (~(pd.isnull(pol_df[['takeoff','azimuth']]))).any(axis=None):
			print('Ignoring the stfile ({}) because takeoff and azimuths are already provided.'.format(p_dict['stfile']))
			p_dict['stfile']=''

	'''
	If using precomputed takeoff/azimuths, discards polarities with takeoffs with large uncertainties.
	If computing takeoff/azimuths, this is handled later in compute_mech().
	'''
	# Discards any measurements with a takeoff uncertainty > pmax
	if not(p_dict['stfile']):
		if p_dict['pmax']>0:
			if 'takeoff_uncertainty' in pol_df.columns:
				drop_index=pol_df[(pol_df['takeoff_uncertainty']>p_dict['pmax'])].index
				if len(drop_index)>0:
					print('Discarded {} polarity measurement(s) with takeoff uncertainties >{} deg'.format(len(drop_index),p_dict['pmax']))
					pol_df=pol_df.drop(drop_index).reset_index(drop=True)

	'''
	Discards earthquakes with fewer than npolmin P-polarities
	'''
	if p_dict['npolmin']>0:
		tmp_df=pol_df.groupby(by='event_id').agg({'p_polarity':'count'})
		discard_event_id=tmp_df[tmp_df['p_polarity']<p_dict['npolmin']].index
		if not(discard_event_id.empty):
			pol_df=pol_df.drop(pol_df[pol_df['event_id'].isin(discard_event_id)].index).reset_index(drop=True)
			print('Discarded {} earthquakes with fewer than {} polarities.'.format(len(discard_event_id),p_dict['npolmin']))

	'''
	Removes cataloged earthquakes that have no selected measurements
	'''
	if len(cat_df):
		cat_consider_flag=~(cat_df['event_id'].isin(pol_df['event_id']))
		if cat_consider_flag.any():
			cat_df=cat_df.drop(cat_df[cat_consider_flag].index).reset_index(drop=True)

	'''
	Reads the velocity model files and creates (or loads) the lookup tables.
	'''
	if p_dict['compute_takeoff_azimuth']:
		lookup_dict=fun.create_lookup_table(p_dict)
	else:
		lookup_dict={'deptab':[],'delttab':[],'table':[]}

	'''
	Sets up array with direction cosines for all coordinate transformations
	'''
	if p_dict['use_fortran']:
		dir_cos_dict={}
	else:
		dir_cos_dict=fun.dir_cos_setup(p_dict)

	'''
	Groups polarities and S/P ratios by event_id
	'''
	group_pol_df=pol_df.groupby(by='event_id')
	event_ids=list(group_pol_df.groups.keys())

	if p_dict['outfile_pol_agree']:
		pol_df['pol_agreement']=0
	if p_dict['outfile_sp_agree']:
		pol_df['sp_diff']=-999.

	# Creates outfile1/outfile2 and adds header lines to them.
	if p_dict['outfile1']:
		out.create_outfile1(p_dict['outfile1'],cat_df,pol_df)
	if p_dict['outfile2']:
		out.create_outfile2(p_dict['outfile2'])

	if not(event_ids):
		print('No mechanisms to compute. Exiting.')
		quit()

	'''
	Computes the focal mechanisms
	'''
	mech_runtime_start=time.time()
	num_events=len(event_ids)
	if p_dict['num_cpus']==1: # Run in serial
		print('Computing mechanisms in serial...')
		for event_x,event_id in enumerate(event_ids):
			mech_dict=compute_mech.compute_mech(event_x,num_events,event_id,group_pol_df.get_group(event_id),
												p_dict,lookup_dict,qual_criteria_dict,cat_df,dir_cos_dict)
			if mech_dict['mech_qual']:
				pol_df.loc[mech_dict['event_index'],'mech_quality']=mech_dict['mech_qual']
				if p_dict['outfile_pol_agree']:
					pol_df.loc[mech_dict['event_index'],'pol_agreement']=mech_dict['pol_agreement_out']
				if p_dict['outfile_sp_agree']:
					pol_df.loc[mech_dict['event_index'],'sp_diff']=mech_dict['sp_diff_out']
				if (p_dict['stfile']) and (p_dict['outfile_pol_info']):
					pol_df.loc[mech_dict['event_index'],'takeoff']=mech_dict['takeoff']
					pol_df.loc[mech_dict['event_index'],'azimuth']=mech_dict['sr_az']
					pol_df.loc[mech_dict['event_index'],'takeoff_uncertainty']=mech_dict['takeoff_uncertainty']
					pol_df.loc[mech_dict['event_index'],'azimuth_uncertainty']=mech_dict['azimuth_uncertainty']
	else: # Run in parallel
		print('Computing mechanisms in parallel...')
		pool=multiprocessing.Pool(processes=p_dict['num_cpus'])

		async_results=[]
		for event_x,event_id in enumerate(event_ids):
			try:
				event_pol_df=group_pol_df.get_group(event_id)
			except:
				print('Error getting parallel result for event_id: {}'.format(event_id))
				continue
			async_results.append(pool.apply_async(compute_mech.compute_mech,
						args=(event_x,num_events,event_id,event_pol_df,p_dict,lookup_dict,qual_criteria_dict,cat_df,dir_cos_dict)))
		pool.close()
		pool.join()

		if any([p_dict['outfile_pol_agree'],p_dict['outfile_sp_agree'],p_dict['outfile_pol_info']]):
			for result in async_results:
				mech_dict=result.get()
				if mech_dict['mech_qual']:
					pol_df.loc[mech_dict['event_index'],'mech_quality']=mech_dict['mech_qual']
					if p_dict['outfile_pol_agree']:
						pol_df.loc[mech_dict['event_index'],'pol_agreement']=mech_dict['pol_agreement_out']
					if p_dict['outfile_sp_agree']:
						pol_df.loc[mech_dict['event_index'],'sp_diff']=mech_dict['sp_diff_out']
					if (p_dict['stfile']) and (p_dict['outfile_pol_info']):
						pol_df.loc[mech_dict['event_index'],'takeoff']=mech_dict['takeoff']
						pol_df.loc[mech_dict['event_index'],'azimuth']=mech_dict['sr_az']
						pol_df.loc[mech_dict['event_index'],'takeoff_uncertainty']=mech_dict['takeoff_uncertainty']
						pol_df.loc[mech_dict['event_index'],'azimuth_uncertainty']=mech_dict['azimuth_uncertainty']

	print('Mech computation runtime: {:.2f} sec'.format(time.time()-mech_runtime_start), flush=True)

	if p_dict['flip_takeoff']:
		pol_df['takeoff']=180-pol_df['takeoff']

	'''
	Determines the polarity agreements at the different stations and writes it to file
	'''
	if p_dict['outfile_pol_agree']:
		out.pol_agree(pol_df,p_dict)

	'''
	Determines the S/P agreements at the different stations and writes it to file
	'''
	if p_dict['outfile_sp_agree']:
		out.sp_agree(pol_df,p_dict)

	'''
	Creates output file with lots of information about the polarity measurements used to compute the focal mechanisms
	'''
	if p_dict['outfile_pol_info']:
		out.pol_info(pol_df,p_dict)

	print('Total runtime: {:.2f} sec'.format(time.time()-total_runtime_start), flush=True)
	sys.exit()

def main():
	SKHASH()

if __name__ == '__main__':
	main()